# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""ui.py includes all graphical user interfaces (mostly console based)
to manage SFLvault elements (customers, machines, services)
"""

import platform
if platform.system() != 'Windows':
    import urwid
    import urwid.raw_display

class DialogExit(Exception):
    pass


class DialogDisplay(object):
    palette = [
        ('body',       'black','light gray','standout'),
        ('border',     'black','dark blue'),
        ('shadow',     'white','black'),
        ('selectable', 'black','dark cyan'),
        ('focus',      'white','dark blue','bold'),
        ('focustext',  'light gray','dark blue'),
        ('editfc',     'white','dark blue','bold'),
        ('editbx',     'light gray','dark blue'),
        ('editcp',     'black','light gray','standout'),
        ]


    def __init__(self, text, height, width, body=None):
        width = int(width)
        if width <= 0:
            width = ('relative', 80)
        height = int(height)
        if height <= 0:
            height = ('relative', 80)

        self.body = body
        if body is None:
            # fill space with nothing
            body = urwid.Filler(urwid.Divider(),'top')

        self.frame = urwid.Frame( body, focus_part='footer')
        if text is not None:
            self.frame.header = urwid.Pile( [urwid.Text(text),
                    urwid.Divider()] )
        w = self.frame

        # pad area around listbox
        w = urwid.Padding(w, ('fixed left',2), ('fixed right',2))
        w = urwid.Filler(w, ('fixed top',1), ('fixed bottom',1))
        w = urwid.AttrWrap(w, 'body')

        # "shadow" effect
        w = urwid.Columns( [w,('fixed', 2, urwid.AttrWrap(
                urwid.Filler(urwid.Text(('border','  ')), "top")
                ,'shadow'))])
        w = urwid.Frame( w, footer = 
                urwid.AttrWrap(urwid.Text(('border','  ')),'shadow'))

        # outermost border area
        w = urwid.Padding(w, 'center', width )
        w = urwid.Filler(w, 'middle', height )
        w = urwid.AttrWrap( w, 'border' )

        self.view = w


    def add_buttons(self, buttons):
        l = []
        for name, exitcode in buttons:
            b = urwid.Button(name, self.button_press)
            b.exitcode = exitcode
            b = urwid.AttrWrap(b, 'selectable', 'focus')
            l.append(b)
        self.buttons = urwid.GridFlow(l, 10, 3, 1, 'center')
        self.frame.footer = urwid.Pile([urwid.Divider(),
                                        self.buttons],
                                       focus_item = 1)

    def button_press(self, button):
        raise DialogExit(button.exitcode)

    def main(self):
        self.ui = urwid.raw_display.Screen()
        self.ui.register_palette( self.palette )
        return self.ui.run_wrapper( self.show_dialog )

    def show_dialog(self):
        self.ui.set_mouse_tracking()
        size = self.ui.get_cols_rows()
        try:
            while True:
                canvas = self.view.render( size, focus=True )
                self.ui.draw_screen( size, canvas )
                keys = None
                while not keys: 
                    keys = self.ui.get_input()
                for k in keys:
                    if urwid.VERSION < (1, 0, 0):
                        check_mouse_event = urwid.is_mouse_event
                    else:
                        check_mouse_event = urwid.util.is_mouse_event
                    if check_mouse_event(k):
                        event, button, col, row = k
                        self.view.mouse_event( size, 
                                event, button, col, row,
                                focus=True)
                    if k == 'window resize':
                        size = self.ui.get_cols_rows()
                    k = self.view.keypress( size, k )

                    if k:
                        self.unhandled_key( size, k)
                        
        except DialogExit as e:
            return self.on_exit( e.args[0] )

    def on_exit(self, exitcode):
        return exitcode, ""

    def unhandled_key(self, size, k):
        if k in ('up','page up'):
            self.frame.set_focus('body')
        if k in ('down','page down'):
            self.frame.set_focus('footer')
        if k == 'enter':
            # pass enter to the "ok" button
            self.frame.set_focus('footer')
            self.buttons.set_focus(0)
            self.view.keypress( size, k )




##
##
## Local dialogs for SFLvault
##
##

class ServiceEditDialogDisplay(DialogDisplay):
    """Edit a service"""
    def __init__(self, data):
        # Temporarily disabled. We'll look into editing right here the groups
        # at anothe moment. We must be able to edit services right now.
        #
        #groups = [urwid.CheckBox('Tag', False, user_data=1),
        #          urwid.CheckBox('tag2', True, user_data=2),
        #          urwid.CheckBox('tag 3', False, user_data=3)]

        if isinstance(data['metadata'], dict):
            metadata_str = '\n'.join('%s=%s' % (key, val) for key, val in list(data['metadata'].items())) + '\n'
        else:
            metadata_str = ''
        inputs = {'url': urwid.Edit("", str(data['url']), wrap='clip'),
                  'machine_id': urwid.Edit("", str(data['machine_id'] or '')),
                  'parent_service_id': urwid.Edit("", str(data['parent_service_id'] or '')),
                  'notes': urwid.Edit("", str(data['notes'])),
                  'metadata': urwid.Edit("", metadata_str, multiline=True)
                  }
                  
                
        
        l = [
            urwid.Columns([('fixed', 15, urwid.Text('URL ', align="right")),
                  urwid.AttrWrap(inputs['url'], 'editbx', 'editfc' )]),
            
            urwid.Columns([('fixed', 15, urwid.Text('Machine ID ',
                                                    align="right")),
                  urwid.AttrWrap(inputs['machine_id'], 'editbx', 'editfc')]),
            
            urwid.Columns([('fixed', 15, urwid.Text('Parent service ID ',
                                                    align="right")),
                  urwid.AttrWrap(inputs['parent_service_id'],
                                 'editbx', 'editfc')]),
            
            urwid.Columns([('fixed', 15, urwid.Text('Notes ', align="right")),
                  urwid.AttrWrap(inputs['notes'], 'editbx','editfc' )]),
            urwid.Divider(),
            urwid.Columns([('fixed', 15, urwid.Text('Metadata ', align="right")),
                           urwid.AttrWrap(inputs['metadata'], 'editbx', 'editfc')])
            
            # Temp disabled (read above)
            #urwid.Divider(),
            #urwid.Columns([('fixed', 15, urwid.Text('Groups ', align="right")),
            #               urwid.AttrWrap(urwid.Pile(groups, 1), 'selectable')])
            ]


        walker = urwid.SimpleListWalker(l)
        lb = urwid.ListBox( walker )
        DialogDisplay.__init__(self, "Edit service", 22, 70, lb)

        self.frame.set_focus('body')
        
        self.add_buttons([("OK", 1), ("Cancel", 0)])

        self.data = data
        self.inputs = inputs

    def run(self):
        """Show the dialog box, and get the results afterwards."""

        exitcode, exitstring = self.main()

        if exitcode:
            data = {}
            # prepare to save
            t = ['parent_service_id', 'machine_id', 'url', 'notes']
            for x in t:
                if self.inputs[x].edit_text != self.data[x]:
                    data[x] = self.inputs[x].edit_text
            metadata = {}
            for line in self.inputs['metadata'].edit_text.split('\n'):
                parts = line.split('=', 1)
                if len(parts) == 2:
                    metadata[parts[0].strip()] = parts[1].strip()
            data['metadata'] = metadata
            return True, data
        else:
            # return nothing..
            return False, None



class MachineEditDialogDisplay(DialogDisplay):
    """Edit a machine"""
    def __init__(self, data):

        # Create fields..
        inputs = {'customer_id': urwid.Edit("", str(data['customer_id'] or ''))}
        for x in ['ip', 'name', 'fqdn', 'location', 'notes']:
            inputs[x] = urwid.Edit("", str(data[x])) # , wrap='clip'
                  
        l = [
            urwid.Columns([('fixed', 15, urwid.Text('Customer ID ',
                                                    align="right")),
                  urwid.AttrWrap(inputs['customer_id'], 'editbx', 'editfc')]),
            
            urwid.Columns([('fixed', 15, urwid.Text('Name ', align="right")),
                  urwid.AttrWrap(inputs['name'], 'editbx', 'editfc' )]),
            
            urwid.Columns([('fixed', 15, urwid.Text('IP ', align="right")),
                  urwid.AttrWrap(inputs['ip'], 'editbx', 'editfc' )]),
            
            urwid.Columns([('fixed', 15, urwid.Text('FQDN ', align="right")),
                  urwid.AttrWrap(inputs['fqdn'], 'editbx', 'editfc' )]),
            
            urwid.Columns([('fixed', 15, urwid.Text('Location ',
                                                    align="right")),
                  urwid.AttrWrap(inputs['location'], 'editbx', 'editfc' )]),
            
            urwid.Columns([('fixed', 15, urwid.Text('Notes ', align="right")),
                  urwid.AttrWrap(inputs['notes'], 'editbx','editfc' )]),
            

            ]


        walker = urwid.SimpleListWalker(l)
        lb = urwid.ListBox( walker )
        DialogDisplay.__init__(self, "Edit machine", 22, 70, lb)

        self.frame.set_focus('body')
        
        self.add_buttons([("OK", 1), ("Cancel", 0)])

        self.data = data
        self.inputs = inputs



    def run(self):
        """Show the dialog box, and get the results afterwards."""

        exitcode, exitstring = self.main()

        if exitcode:
            data = {}
            # prepare to save
            t = ['customer_id', 'ip', 'name', 'fqdn', 'location', 'notes']
            for x in t:
                if self.inputs[x].edit_text != self.data[x]:
                    data[x] = self.inputs[x].edit_text

            return True, data
        else:
            # return nothing..
            return False, None




class CustomerEditDialogDisplay(DialogDisplay):
    """Edit a customer"""
    def __init__(self, data):

        # Create fields..
        inputs = {}
        for x in ['name']:
            inputs[x] = urwid.Edit("", str(data[x])) # , wrap='clip'
                  
        l = [
            urwid.Columns([('fixed', 15, urwid.Text('Name ',
                                                    align="right")),
                           urwid.AttrWrap(inputs['name'], 'editbx', 'editfc')]),
            ]


        walker = urwid.SimpleListWalker(l)
        lb = urwid.ListBox( walker )
        DialogDisplay.__init__(self, "Edit customer", 22, 70, lb)

        self.frame.set_focus('body')
        
        self.add_buttons([("OK", 1), ("Cancel", 0)])

        self.data = data
        self.inputs = inputs



    def run(self):
        """Show the dialog box, and get the results afterwards."""

        exitcode, exitstring = self.main()

        if exitcode:
            data = {}
            # prepare to save
            t = ['name']
            for x in t:
                if self.inputs[x].edit_text != self.data[x]:
                    data[x] = self.inputs[x].edit_text

            return True, data
        else:
            # return nothing..
            return False, None


class GroupEditDialogDisplay(DialogDisplay):
    """Edit a group"""
    def __init__(self, data):

        # Create fields..
        inputs = {}
        # TODO: Eventually add 'hidden' to tweak hidden status.
        for x in ['name']:
            inputs[x] = urwid.Edit("", str(data[x])) # , wrap='clip'
                  
        l = [
            urwid.Columns([('fixed', 15, urwid.Text('Name ',
                                                    align="right")),
                           urwid.AttrWrap(inputs['name'], 'editbx', 'editfc')]),
            ]


        walker = urwid.SimpleListWalker(l)
        lb = urwid.ListBox( walker )
        DialogDisplay.__init__(self, "Edit group", 22, 70, lb)

        self.frame.set_focus('body')
        
        self.add_buttons([("OK", 1), ("Cancel", 0)])

        self.data = data
        self.inputs = inputs



    def run(self):
        """Show the dialog box, and get the results afterwards."""

        exitcode, exitstring = self.main()

        if exitcode:
            data = {}
            # prepare to save
            t = ['name']
            for x in t:
                if self.inputs[x].edit_text != self.data[x]:
                    data[x] = self.inputs[x].edit_text

            return True, data
        else:
            # return nothing..
            return False, None




# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008-2009  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from sflvault.client.remoting import *
from sflvault.client.utils import *
from sflvault.client.fallback import SFLvaultFallback
import struct, fcntl, termios, signal # for term size signaling..
import optparse
import random
import pipes
import time
import sys
import os
import platform
if platform.system() != 'Windows':
    import pexpect

# Enum used in the services' `provides`
PROV_PORT_FORWARD = 'PROV_PORT_FORWARD'
PROV_SHELL_ACCESS = 'PROV_SHELL_ACCESS'
PROV_MYSQL_CONSOLE = 'PROV_MYSQL_CONSOLE'
PROV_POSTGRES_CONSOLE = 'PROV_POSTGRES_CONSOLE'

# Operational modes
OP_DIRECT = 'OP_DIRECT'                # - When, let's say 'ssh' is going to
                                       # spawn a process.
OP_THRGH_FWD = 'OP_THRGH_FWD'          # - When we must go through a port
                                       # that's been forwarded.
                                       # In that case, the parent must provide
                                       # with  forwarded host and port.
OP_THRGH_SHELL = 'OP_THRGH_SHELL'      # - When a command must be sent through
                                       # an open shell



### Service definitions

class ExpectClass(object):
    def __init__(self, service_obj, strings_and_funcs=None):
        # Hold the parent Service object here..
        self.service = service_obj
        self.error = None

        # Quick references..
        self.cnx = self.service.shell_handle

        funcs = []
        strings = []
        if strings_and_funcs:
            for s, f in strings_and_funcs:
                strings.append(s)
                funcs.append(getattr(self, f))
        for x in dir(self):
            if x.startswith('_'):
                continue
            
            func = getattr(self, x)

            if callable(func) and func.__doc__:
                funcs.append(func)
                strings.append(func.__doc__)

        idx = self._expect(strings)

        funcs[idx]()

    def _expect(self, strings):
        """Wrap the expect() call"""
        timeout = self.service.timeout

        try:
            idx = self.cnx.expect(strings, timeout=timeout)
        except pexpect.TIMEOUT as e:
            self._flush()
            if hasattr(self, '_timeout'):
                self._timeout()
            else:
                sys.stdout.write("\nTimed out")

            self.error = "Timed out"
            raise ServiceExpectError("Timed out")

        self._flush()

        return idx
    
    def _flush(self):
        sys.stdout.write(self.cnx.before)
        sys.stdout.write(self.cnx.after)
        sys.stdout.flush()


class ExpectShell(ExpectClass):
    """When initialized, this class will wait for a shell prompt via pexpect
    and call its ``shell`` method when it sees one.

    The shell prompt can be configured for each service using the
    ``shell_prompt`` metadata key.
    """
    def __init__(self, service, strings_and_funcs=None):
        strings_and_funcs = strings_and_funcs[:] if strings_and_funcs else []
        if isinstance(service.data['metadata'], dict):
            prompt = service.data['metadata'].get('prompt') or r'[^ ]*@.*:.*[$#] '
        else:
            prompt = r'[^ ]*@.*:.*[$#] '
        strings_and_funcs.append((prompt, 'shell'))
        ExpectClass.__init__(self, service, strings_and_funcs)

    def shell(self):
        pass


class ShellService(Service):
    """Abstract class for services running over ssh. Helper funcs."""


    def __init__(self, data, timeout=45):
        # Call parent
        Service.__init__(self, data, timeout)
        
    def interact(self):

        # To grab the window changing signals
        def sigwinch_passthrough (sig, data):
            s = struct.pack("HHHH", 0, 0, 0, 0)
            a = struct.unpack('hhhh', fcntl.ioctl(sys.stdout.fileno(),
                                                  termios.TIOCGWINSZ , s))
            self.shell_handle.setwinsize(a[0],a[1])

        # Map window resizing signal to function
        signal.signal(signal.SIGWINCH, sigwinch_passthrough)

        # Call USER's terminal and get term size.
        s = struct.pack("HHHH", 0, 0, 0, 0)
        fd_stdout = sys.stdout.fileno()
        x = fcntl.ioctl(fd_stdout, termios.TIOCGWINSZ, s)
        sz = struct.unpack("HHHH", x)
        # Send to the remote terminals..
        self.shell_handle.setwinsize(sz[0], sz[1])


        try:
            # Go ahead, you're free to go !
            while True:
                self.shell_handle.interact(escape_character=sflvault_escape_chr)
                # Drop in the shell, only if connection is still open
                if not self.shell_handle.isalive():
                    break

                s = SFLvaultFallback(self.chain, self.shell_handle)
                try:
                    s._run()
                except (KeyboardInterrupt, EOFError) as e:
                    s._return_completer()
                continue

                # TODO we need to check if shell is still alive

            # Reset signal
            signal.signal(signal.SIGWINCH, signal.SIG_DFL)

            print("[SFLvault] Escaped from %s, calling postwork." % self.__class__.__name__)
        except OSError as e:
            print("[SFLvault] %s disconnected: %s" % (self.__class__.__name__,
                                                      e))
            raise e


    def postwork(self):
        # Close any open connection.
        if not self.shell_handle.closed:
            print("Closing connection for service %s" % self)
            self.shell_handle.close()
        # NOTE: this will close all the `ssh` tunnel when it was continued
        #       through shells



class ssh(ShellService):
    """ssh protocol service handler"""

    def __init__(self, data, timeout=45):
        # Call parent
        ShellService.__init__(self, data, timeout)
        self.provides_modes = set([PROV_SHELL_ACCESS, PROV_PORT_FORWARD])
        self.local_forwards = []
        self.remote_forwards = []
        # TODO: Add SOCKS forward...
        
    ssh_pki_auth = False   # Overwridden by ssh_pki

    def has_forwards(self):
        """Returns True if we have something to forward"""
        return bool(self.local_forwards) or bool(self.remote_forwards)

    def optparse(self, command_line):
        """Parse the command line for port-forwarding options"""
        parser = optparse.OptionParser(usage=optparse.SUPPRESS_USAGE)
        parser.set_usage('... [-L port:hostname:port [-L ...]] ' \
                         '[-R port:hostname:port [-R ...]]')
        parser.add_option('-L', dest='locals', action='append', type='string',
                          default=[],
                          help="Listen locally, then forward (see man ssh)")
        parser.add_option('-R', dest='remotes', action='append', type='string',
                          default=[],
                          help="Listen remotely, forward locally. ")
        opts, args = parser.parse_args(args=command_line)
        # We shouldn't have remaining arguments
        if args:
            raise RemotingError('Unknown parameters: %s' % args)
        # Save locals and remotes
        def parse_forward(fwd):
            # Return tuples in the form: (bind_address, port, host, hostport)
            el = tuple(fwd.split(':'))
            if len(el) == 3:
                return ('localhost',) + el
            elif len(el) == 4:
                return el
            else:
                raise RemotingError('Parameters to -L or -R must be in the form: [bind_address:]port:host:hostport')

        for l in opts.locals:
            #print "Requesting LOCAL forward for: %s" % l
            self.chain.add_port_forward(None, 'local', *parse_forward(l))
        for r in opts.remotes:
            #print "Requesting REMOTE forward for: %s" % r
            self.chain.add_port_forward(None, 'remote', *parse_forward(r))
        #self.local_forwards = [parse_forward(l) for l in opts.locals]
        #self.remote_forwards = [parse_forward(r) for r in opts.remotes]

    def required(self, req=None):
        """Verify if this service handler can provide `req` type of
        connection.

        Child can ask for SHELL_ACCESS or PORT_FORWARD to be provided. It
        can also ask for nothing. It is the case of the last child, which
        has required() invoked with mode=None, since there's no other
        child that requires something.

        When mode=None, plugin is most probably the last child, and the
        end to which we want to connect.
        """

        if not self.child and not self.parent:
            # We're alone in the world, we've setting up DIRECT access (spawn a
            # process)
            self.operation_mode = OP_DIRECT
            return True

        # Can we provide what you need ?
        if not self.provides(req):
            # Sorry, we don't know what you're talking about :)
            raise ServiceRequireError("ssh module can't provide '%s'" % req)

        if self.ssh_pki_auth and self.parent:
            # If the command line asked it, or we're authenticating through
            # ssh keys, try to make port-forwards.
            if self.parent:
                self.operation_mode = OP_THRGH_FWD
                # Setup port-forwards
                randport = random.randint(58000, 60000)
                self.chain.add_port_forward(self, 'local', 'localhost',
                                            randport, self.url.hostname,
                                            self.url.port or 22)
                self.op_through_host = 'localhost'
                self.op_through_port = randport
            else:
                self.operation_mode = OP_DIRECT

            if self.parent:
                return self.parent.required(PROV_PORT_FORWARD)
            else:
                return True

        if not self.parent:
            # If there's no parent, then we're going to spawn ourself.
            self.operation_mode = OP_DIRECT
        else:
            # Otherwise, 'ssh' must be called from within a shell.
            self.operation_mode = OP_THRGH_SHELL
            # Ask for a PORT_FORWARD if that's what was asked of us,
            # otherwise, just give us a normal shell.
            return self.parent.required(req or PROV_SHELL_ACCESS)

        return True


    def prework(self):

        # Prework classes
        class expect_shell(ExpectShell):

            def terminal_type(self):
                "Terminal type\?.*$"
                sys.stdout.write(" [sending: xterm] ")
                self.cnx.sendline("xterm")
                sys.stdout.flush()
            
            def shell(self):
                pass # We're in :)
            
            def denied(self):
                'Permission denied.*$'
                self.cnx.sendcontrol('c')
                raise ServiceExpectError("Failed to authenticate")
            
            def failed_login(self):
                'assword:'
                self.cnx.sendcontrol('c')
                print("")
                raise ServiceExpectError("Failed to authenticate")

            def are_you_sure(self):
                "(?i)are you sure you want to continue connecting.*\? "
                ans = input()
                self.cnx.sendline(ans)
                expect_shell(self.service)

            def forward_error(self):
                "Privileged ports can only be forwarded by root"
                print("")
                raise ServiceExpectError("Port forward failed.")

        class expect_login(ExpectClass): 
            #forward_error = expect_shell.forward_error

            def sendlogin(self):
                'assword:'
                sys.stdout.write(" [sending password...] ")
                self.cnx.sendline(self.service.data['plaintext'])
                sys.stdout.flush()

                expect_shell(self.service)

            def werein(self):
                'Last login'
                print(" [We're in! (using shared-key?)] ", end=' ')

            def are_you_sure(self):
                "(?i)are you sure you want to continue connecting.*\? "
                # TODO: are you always sure ??
                ans = input()
                self.cnx.sendline(ans)

                expect_login(self.service)

            def remote_host_changed(self):
                "REMOTE HOST IDENTIFICATION HAS CHANGED!"

                ans = input("Would you like SFLvault to remove that offending line and try again (yes/no)? ")
                if ans.lower() == 'yes' or ans.lower() == 'y':
                    idx = self._expect([r"Offending key in ([^:]+):(\d+)"])
                    cmd = 'sed -i %sd "%s"' % (self.cnx.match.group(2),
                                               self.cnx.match.group(1))

                    if self.service.operation_mode in [OP_DIRECT,
                                                       OP_THRGH_FWD]:
                        # Remove the line
                        os.system(cmd)
                    else: # operation_mode == OP_THRGH_SHELL
                        self.cnx.sendline(cmd)

                    raise ServiceExpectError("SFLvault removed offending key. Please try again.")
                else:
                    raise ServiceExpectError("Remote host identification has changed. Please resolve problem.")

                self._expect([pexpect.EOF])
            

        # Default user:
        user = self.url.username or 'root'

        sshcmd = "ssh"
        if self.operation_mode in [OP_DIRECT, OP_THRGH_SHELL]:
            print("\n... Trying to login to %s as %s ..." % (self.url.hostname,
                                                             user))
            sshcmd += " -l %s %s" % (user, self.url.hostname)
            if self.url.port:
                sshcmd += " -p %d" % (self.url.port)
        elif self.operation_mode == OP_THRGH_FWD:
            # Take the bound port on the first (OP_DIRECT) ssh connection.
            port = self.op_through_port
            host = self.op_through_host
            print("... Trying to login to %s (through local port-forward, " \
                  "port %d) as %s ..." % (self.url.hostname, port, user))
            # TODO: Do not take into consideration the `known hosts` list.
            sshcmd += " -l %s %s" % (user,host)
            sshcmd += " -p %d" % (port)
            sshcmd += " -o NoHostAuthenticationForLocalhost=yes"
            #time.sleep(0)

        # Check if we have private key support
        pki_cmd = None
        if self.ssh_pki_auth:
            # Write the private-key to a temp file.
            randfifo = '/tmp/sshpki%d' % random.randint(100000,999999)
            secret = pipes.quote(self.data['plaintext'])
            pki_cmd = "rm -f %s ; touch %s ; chmod 0600 %s ; " \
                      "((echo %s > %s; sleep 4; rm %s) &);" % (randfifo,
                                                               randfifo,
                                                               randfifo,
                                                               secret, randfifo,
                                                               randfifo)
            sshcmd += " -i %s " % randfifo

        #if PROV_PORT_FORWARD in self.provides_modes:
        if self.has_forwards():
            # Do the last forwarding
            sshcmd += " %s %s " % (' '.join(self.local_forwards),
                                   ' '.join(self.remote_forwards))

        if self.operation_mode in [OP_DIRECT, OP_THRGH_FWD]:
            if pki_cmd:
                print("Running SSH+PKI setup...")
                os.system(pki_cmd)
            print("EXECUTING: %s" % sshcmd)
            cnx = pexpect.spawn(sshcmd, encoding='utf-8')
        elif self.operation_mode == OP_THRGH_SHELL:
            cnx = self.parent.shell_handle
            if pki_cmd:
                print("Sending SSH+PKI setup...")
                cnx.sendline(pki_cmd)
            print("SENDING LINE: %s" % sshcmd)
            cnx.sendline(sshcmd)
        else:
            # NOTE: This should be dead code:
            raise RemotingError("No way to go through there. Woah? How did we get here ?")

        self.shell_handle = cnx

        if self.ssh_pki_auth:
            expect_shell(self)
        else:
            expect_login(self)





class ssh_pki(ssh):
    """SSH with private-key handler"""
    ssh_pki_auth = True   # Overwridden by ssh_pki

    @staticmethod
    def ask_password(edit, parsed_url):
        """Function to grab password"""
        print("""Copy and paste the SSH PRIVATE KEY, which looks like:
-----BEGIN (D|R)SA PRIVATE KEY-----
B64DEADBEEF...==
-----END (D|R)SA PRIVATE KEY-----
""")
        pk = []
        while True:
            chunk = input(">>> ")
            pk.append(chunk)
            if '---END' in chunk:
                break
        return '\n'.join(pk)
        


class content(Service):
    """Static content store.  Will store any blob into the database as the secret"""
    @staticmethod
    def ask_password(edit, parsed_url):
        """Function to grab the content and set it as the secret"""
        print("""Paste the content to store, and hit Ctrl+C""")
        pk = []
        try:
            while True:
                chunk = input(">>> ")
                pk.append(chunk)
        except KeyboardInterrupt as e:
            print("")
            return '\n'.join(pk)

class sflvault(content):
    """Inherit password entry, when storing other Vault's Private key"""
    pass



# TODO: document that the `sudo` doesn't need a password, it takes the password
# from the parent service.
class sudo(ShellService):
    """sudo app. service handler"""

    provides_modes = set([PROV_SHELL_ACCESS])

    def required(self, req=None):
        """Check requirements for sudo

        See ssh service handle for documentation of `required`
        """

        # We must be over an 'ssh' to do anything!
        if not self.parent or not self.parent.required(PROV_SHELL_ACCESS):
            raise ServiceRequireError("`sudo` must be child of a shell")

        # We can provide SHELL
        if not self.provides(req):
            raise ServiceRequireError("`sudo` module can't provide '%s'" % req)

        # TODO: pass-through PORT_FORWARD provisioning if child supports it..
        return True


    def prework(self):
        # Inherit shell handle
        self.shell_handle = self.parent.shell_handle

        class expect_waitshell(ExpectShell):

            # def shell(self): from ExpectShell (uses prompt)

            def failed(self):
                'assword:'
                self.cnx.sendintr()
                print("")
                raise ServiceExpectError("Failed to authenticate sudo://")

            def failed2(self):
                'Sorry'
                self.failed()

        class expect_sudowork(ExpectShell):

            # def shell(self): from ExpectShell (uses prompt)

            def sendpass(self):
                'assword:'

                sys.stdout.write(" [sending password...] ")
                self.cnx.sendline(self.service.parent.data['plaintext'])

                expect_waitshell(self.service)

            def sendpass2(self):
                r'assword for.*:'
                self.sendpass()

        # Send command
        self.shell_handle.sendline('sudo -i')

        expect_sudowork(self)
        
    # Call parent
    #def interact(self):
    # Call parent
    #def postwork(self):

   
class su(ShellService):
    """su app. service handler"""

    provides_modes = set([PROV_SHELL_ACCESS])

    def required(self, req=None):
        """Check requirements for su
        
        See ssh service handle for documentation of `required`
        """

        # We must be over an 'ssh' to do anything!
        if not self.parent or not self.parent.required(PROV_SHELL_ACCESS):
            raise ServiceRequireError("`su` must be child of a shell")

        # We can provide SHELL
        if not self.provides(req):
            raise ServiceRequireError("`su` module can't provide "\
                                      "'%s'" % req)

        # TODO: pass-through PORT_FORWARD provisioning if child supports it..
        return True


    def prework(self):
        # Inherit shell handle
        self.shell_handle = self.parent.shell_handle

        class expect_waitshell(ExpectShell):

            def shell(self):
                return True

            def failed(self):
                'assword:'
                raise ServiceExpectError("Failed to authenticate su://")

            def failed2(self):
                'incorrect'
                self.failed()

        class expect_suwork(ExpectShell):
            def sendpass(self):
                'assword:'

                sys.stdout.write(" [sending password...] ")
                self.cnx.sendline(self.service.data['plaintext'])

                expect_waitshell(self.service)

            def shell(self):
                return True
                

        # Send command
        self.shell_handle.sendline('su %s' % self.url.username or 'root')

        expect_suwork(self)
        
    # Call parent
    #def interact(self):
    # Call parent
    #def postwork(self):



class mysql(ShellService):
    """mysql app. service handler"""

    provides_modes = set([PROV_MYSQL_CONSOLE])
       
    def required(self, req=None):
        """Verify provided modes.
        
        See ssh service handle for documentation of `required`
        """

        if self.provides(req):
            self.provides_modes = set([req])
        else:
            # Sorry, we don't know what you're talking about :)
            raise ServiceRequireError("mysql module can't provide '%s'" % req)
        
        if not self.parent:
            raise ServiceRequireError("mysql service can't be executed locally, it has to be behind a SHELL_ACCESS-providing module.")

        self.operation_mode = OP_THRGH_SHELL
        return self.parent.required(PROV_SHELL_ACCESS)


    def prework(self):

        class expect_mysql_shell(ExpectClass):
            def shell(self):
                'mysql>'
                pass # We're in :)

            def error(self):
                'ERROR \d*'
                raise ServiceExpectError('Failed to authenticate with mysql')

        class expect_mysql(ExpectClass):
            def login(self):
                'assword:'
                sys.stdout.write(" [sending password...] ")
                self.cnx.sendline(self.service.data['plaintext'])

                expect_mysql_shell(self.service)

            def error(self):
                'ERROR \d*'
                raise ServiceExpectError("Failed authentication for mysql "\
                                         "at program launch")
            
            def notfound(self):
                'command not found.*$'
                raise ServiceExpectError('mysql client not installed on server')


        # Bring over here the parent's shell handle.
        cnx = self.parent.shell_handle
        self.shell_handle = cnx

        # Select a database directly ?
        db = ''
        if self.url.path:
            if self.url.path != '/':
                db = self.url.path.lstrip('/')
                
        cmd = "mysql -h %s -u %s -p %s" % (self.url.hostname, self.url.username,
                                           db)

        cnx.sendline(cmd)


        expect_mysql(self)


    # interact() and postwork() inherited
    

class postgres(ShellService):
    """PostgreSQL service handler"""

    provides_modes = set([PROV_POSTGRES_CONSOLE])
       
    def required(self, req=None):
        """Verify provided modes.
        
        See ssh service handle for documentation of `required`
        """

        if self.provides(req):
            self.provides_modes = set([req])
        else:
            # Sorry, we don't know what you're talking about :)
            raise ServiceRequireError("postgres module can't provide '%s'" % req)
        
        if not self.parent:
            raise ServiceRequireError("postgres service can't be executed locally, it has to be behind a SHELL_ACCESS-providing module.")

        self.operation_mode = OP_THRGH_SHELL
        return self.parent.required(PROV_SHELL_ACCESS)


    def prework(self):

        class expect_postgres_shell(ExpectClass):
            def shell(self):
                '\w*=#'
                pass # We're in :)

            def error(self):
                'postgres: FATAL:  password authentication failed \w* "\w*"'
                raise ServiceExpectError('Failed to authenticate with postgres')

        class expect_postgres(ExpectClass):
            def login(self):
                'assword for user \w*:'
                sys.stdout.write(" [sending password...] ")
                self.cnx.sendline(self.service.data['plaintext'])

                expect_postgres_shell(self.service)

            def error(self):
                'postgres: FATAL:  password authentication failed \w* "\w*"'
                raise ServiceExpectError("Failed authentication for postgres "\
                                         "at program launch")
            
            def notfound(self):
                'command not found.*$'
                raise ServiceExpectError('postgres client not installed on server')


        # Bring over here the parent's shell handle.
        cnx = self.parent.shell_handle
        self.shell_handle = cnx

        cmd = "psql -U %s -W -h %s" % (self.url.username, self.url.hostname)

        # Select a port  directly ?
        if int(self.url.port):
            cmd += " -p %s" % self.url.port

        # Select a database directly ?
        db = ''
        if self.url.path:
                if self.url.path != '/':
                    db = self.url.path.lstrip('/')
                    cmd += " -d %s" % db

        cnx.sendline(cmd)

        expect_postgres(self)

    # interact() and postwork() inherited


# TODO: do something :) launch vncviewer! Require a port-forward
# if behind some `ssh`.
class vnc(Service):
    """vnc protocol service handler"""

    # Modes this service handler provides
    provides_modes = set([])




# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This module contains utilities for the remoting feature of SFLvault.

The base Service class and Chain class are defined. Those are the basic
service-chaining machinery. They are generic.

For individual services, like `ssh` and `http`, see services.py.

Entry-points taken into account here:

  [sflvault.services]

must list some instances of the `Service` class or some derived class.

"""

from sflvault.client.utils import *

import random
import sys
import platform
if platform.system() != 'Windows':
    import pexpect
import urllib.parse

__all__ = ['Service', 'Chain']

# Mother class for service handlers.
class Service(object):
    """Object to describe a service (ex: ssh) and check how it is possible
    for it to establish a connection, even through a chain of services (like
    other ssh's, vpns, port-forwards, etc..)

    his class is a mother class to be inherited from all the service-specific
    objects (ssh, vnc, rdp, mysql, http, https, drac, ftp, postgresql, su,
    sudo, etc..)

    NOTE: Always go through this function first when overloading this class.

    """
    def __init__(self, data, timeout=30, command_line=''):
        self.data = data
        self.url = urllib.parse.urlparse(data['url'])
        self.timeout = timeout
        self.command_line = command_line
        # We need a copy, as the modes can change when configuring chain.
        self.provides_modes = set() # List what a service handler supports
               # (gives SHELL access or can provide PORT_FORWARD, etc. Defaults
               # to nothing special: plugin is always an end-point.
        self.operation_mode = None # By default, no operation mode, until we
                                   # define one in required() for each service.
        # If this service is to go through another host/port
        self.op_through_host = None
        self.op_through_port = None
        # Used for link-lists
        self.parent = None
        self.child = None
        # This will be filled for child requiring a `shell` handle
        self.shell_handle = None

    def __repr__(self):
        return "<%s: %s>" % (self.__class__.__name__,
                             self.data.get('url'))

    def set_child(self, child):
        """Set the child of this service (sets the parent on the child too)"""
        self.child = child
        child.parent = self
        
    def unlink_child(self):
        """Unlink parent from child.

        Call on the first parent to unlink all elements of the chain.

        WARN: This should be called to avoid circular references.
        """
        c = self.child
        p = self.parent

        self.child = None
        self.parent = None
        self.chain = None
        
        if c:
            c.unlink()
        # Not needed (loops in vain) if called from the first parent:
        #if p:
        #    p.unlink()


    def provides(self, mode):
        """Return True of False, whether it supports the provisioning mode"""
        if not mode:
            return True
        elif isinstance(mode, (set, list)):
            return set(mode).issubset(self.provides_modes)
        else:
            return mode in self.provides_modes


    # Abstract functions that must exist in childs
    def required(self, req=None):
        """Called to setup the chain and make dependencies/providings checks"""
        raise NotImplementedError

    def prework(self):
        """Called when Chain::connect() is called on each Service() to setup
        communication for that particular service."""
        raise NotImplementedError
        
    def interact(self):
        """This function will be called only if the last of the chain"""
        raise NotImplementedError
        
    def postwork(self):
        """Called after interact() is finished for the last child, in reverse order."""
        raise NotImplementedError

    ### Optional methods on derived classes (see services.ssh)
    #
    # def optparse(self, command_line):
    #     pass
    #
    # @staticmethod
    # def ask_password(edit:bool, url:string):
    #     pass

class Chain(object):
    """Service chain handler.

    Usage:

    >>> chain = Chain(services_struct)
    # Setup the chain, make sure everything is fine.
    >>> chain.setup()
    >>> chain.connect()

    services_struct is the structure received from an sflvault.show() XML-RPC
    call."""

    def __init__(self, services=None, command_line=''):
        """Initialise the Chain object"""
        self.services = services
        self.command_line = command_line
        # Make sure you call setup() once.
        self.ready = False
        # Link-list of the Service objects, 
        self.service_list = None


    def debug_chain(self):
        print("self.ready: ", self.ready)
        print("self.service_list:")
        for s in self.service_list:
            print(" ", s)
            print("    self.data['url']: ", s.data['url'])
            print("    self.provides_modes: ", s.provides_modes)
            print("    self.local_forwards: ", s.local_forwards)
            print("    self.remote_forwards: ", s.remote_forwards)
            print("    self.ssh_pki_auth: ", s.ssh_pki_auth)
            print("    self.has_forwards(): ", s.has_forwards())
            print("    self.child: ", s.child)
            print("    self.parent: ", s.parent)
            print("    self.operation_mode: ", s.operation_mode)
            print("    self.op_through_port: ", s.op_through_port)
            print("    self.op_through_host: ", s.op_through_host)



    def add_port_forward(self, service, direction, src_host, src_port,
                         dest_host, dest_port):
        """Add port forwards to every one in the chain starting from under
        `service`.

        This will generate intermediate ports to make sure you have the required
        port on the end-point.

        """
        if direction == 'remote':
            dest_host, src_host = src_host, dest_host
            dest_port, src_port = src_port, dest_port
        first_src = src_host, src_port
        dest = dest_host, dest_port
        lst = self.service_list
        # All services or the one under the specified `service`
        #print "List, from service: %s" % service
        for srv in reversed(lst if service == None else lst[:lst.index(service)]):
            # Last in chain ?
            if not srv.parent:
                src = first_src
            else:
                src = 'localhost', random.randint(58000, 60000)
            if direction == 'remote':
                add = "-R %s:%s:%s" % (dest[1], src[0], src[1])
                #print "ADD to %s: %s" % (srv, add)
                srv.remote_forwards.append(add)
            elif direction == 'local':
                add = "-L %s:%s:%s:%s" % (src + dest)
                #print "ADD to %s: %s" % (srv, add)
                srv.local_forwards.append(add)
            dest = src

        
    def setup(self):
        """Tries to set up the service chain (one service at a time).
        
        Raises exceptions if a service can't be handled, or doesn't
        have all the required connection types (ex: no port forward
        exist prior to accessing a service which requires one, just like
        you can't establish an SSH connection over an HTTP connection)."""


        # Create Service objects for each of the service in the hierarchy.
        service_list = []
        for srvdata in self.services:
            parsed_url = urllib.parse.urlparse(srvdata['url'])
            service = None

            for ep in services_entry_points():
                if ep.name == parsed_url.scheme:
                    srvobj = ep.load()
                    service = srvobj(srvdata)
                    service.chain = self
                    break

            if not service:
                raise RemotingError("Service %s has no handler" % srvdata['url'])

            service_list.append(service)
            
        # Link the services as parent/child.
        for i in range(len(service_list) - 1):
            service_list[i].set_child(service_list[i+1])

        self.service_list = service_list

        # Dispatch the command_line, to the latest handler that supports it.
        done = False
        for serv in reversed(service_list):
            if hasattr(serv, 'optparse'):
                # TODO: deal with optparse errors (try), and display good errors
                serv.optparse(self.command_line)
                done = True
                break
        if not done and self.command_line:
            raise RemotingError("No service handle to take care of your command line arguments: %s" % self.comment_line)

        # Call .required() on the LAST element, and setup all the chain.
        if not self.service_list[-1].required():
            raise RemotingError("No services route to destination")
        else:
            self.ready = True
            return True


    def unlink_all(self):
        # Is this used anywhere ?
        sl = service_list
        self.service_list = None
        sl[0].unlink_child()
        

    def connect(self):
        """Connect to the servers in cascade, and do whatever is possible
        to handle connectivity of any time (ssh connections, port forwards,
        etc)"""
        if not self.ready:
            raise RemotingException("Chain not ready. Call setup() first.")

        last = None
        go_break = False
        active_services = []
        # Setup communication
        for srv in self.service_list:
            try:
                srv.prework()
            except ServiceExpectError as e:
                # Show error, fall back, and interact to last if it exists.
                print("[SFLvault] Error connecting to %s: %s" % (srv.data['url'], str(e)))
                break
            except TypeError as e:
                print("[SFLvault] Timed out.")
                sys.stdout.write(srv.shell_handle.before)
                go_break = True
            except pexpect.EOF:
                print("[SFLvault] Child exited. Following is the output of the command: %s" % (
                    srv.shell_handle.before
                ))
                break
            except KeyboardInterrupt as e:
                print("")
                print("[SFLvault] Keyboard interrupt, dropping in interactive mode")
                sys.stdout.write(srv.shell_handle.before)
                last = srv
                go_break = True

            active_services.append(srv)
            last = srv

            # In case we want to be dropped in the *latest* shell, the one
            # that didn't succeed.
            if go_break:
                break


        # Interact with end-point
        if last:
            current = last
            while True:
                try:
                    current.interact()
                except OSError as e:
                    # Make suer you clean-up anyways :)
                    break
                except ServiceSwitchException as e:
                    # A way to switch shells..
                    current = e.service
                    continue
                break

        # Close-up communication
        for srv in reversed(active_services):
            srv.postwork()

        #print "Chain::connect() done"



# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008-2009  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pkg_resources as pkgres
import optparse
import os
import urllib
import sys
import xmlrpc.client
import getpass
import shlex
import socket
import platform
if platform.system() != 'Windows':
    import readline

from Crypto.PublicKey import ElGamal
from base64 import b64decode, b64encode
from datetime import *

from sflvault.client.client import SFLvaultClient
from sflvault.common.crypto import *
from sflvault.common import VaultError
from sflvault.client.utils import *
from sflvault.client import ui

class SFLvaultParserError(Exception):
    """For invalid options on the command line"""
    pass


class ExitParserException(Exception):
    """Tells when the parser showed the help for a command."""
    pass

class NoExitParser(optparse.OptionParser):
    """Simple overriding of error handling, so that no sys.exit() is being
    called

    Reference: http://bugs.python.org/issue3079
    """
    def exit(self, status=0, msg=None):
        if msg:
            sys.stderr.write(msg)
        raise ExitParserException()

    def error(self, msg):
        """error(msg : string)

        Print a usage message incorporating 'msg' to stderr and exit.
        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        self.print_usage(sys.stderr)

class SFLvaultShell(object):
    def __init__(self, config, vault=None):
        self.config = config
        self.vault = (vault or SFLvaultClient(config, shell=True))

    def _run(self):
        """Go through all commands on a pseudo-shell, and execute them,
        caching the passphrase at some point."""
        
        print("Welcome to SFLvault. Type 'help' for help.")
        prompt = "SFLvault> "
        
        while True:
            cmd = input(prompt)
            if not cmd:
                continue
            
            # Get sys.argv-like parameters
            args = shlex.split(cmd)

            # Local (shell) cmds take precedence over SFLvaultCommand cmds.
            if len(args) and hasattr(self, args[0]):
                getattr(self, args[0])()
            else:
                parser = NoExitParser(usage=optparse.SUPPRESS_USAGE)
                runcmd = SFLvaultCommand(self.config, self.vault, parser)
                try:
                    runcmd._run(args)
                except ExitParserException as e:
                    pass
                
                if hasattr(runcmd, 'next_command') \
                          and platform.system() != 'Windows':
                    print("[Added to shell history: %s]" % runcmd.next_command)
                    readline.add_history(runcmd.next_command)

    def quit(self):
        """Quit command, only available in the shell"""
        raise KeyboardInterrupt()

    def exit(self):
        """Exit command, only available in the shell"""
        raise KeyboardInterrupt()


class SFLvaultCommand(object):
    """Parse command line arguments, and call SFLvault commands
    on them.

    Each method of this object are SFLvault commands.  They are called when
    you run ``sflvault connect s#1`` on the command line, or when you run
    ``connect s#1`` from within the shell.
    """
    def __init__(self, config=None, vault=None, parser=None):
        """Create a SFLvaultCommand object

        :param config: config filename to use, required if no vault specified
        :param vault: an existing SFLvaultClient object, otherwise it will be created, using specified config
        :param parser: an option parser, otherwise it will be created (recommended)
        """
        self.parser = (parser or optparse.OptionParser(usage=optparse.SUPPRESS_USAGE))
        
        if not config and not vault:
            raise ValueError("`config` required if `vault` not specified")

        # Use the specified, or create a new one.
        self.vault = (vault or SFLvaultClient(config))

    def _run(self, argv):
        """Run a certain command"""
        self.argv = argv     # Bump the first (command name)
        self.args = []       # Used after a call to _parse()
        self.opts = object() #  idem.

        # Setup default action = help
        action = 'help'
        self.listcmds = False
        if len(self.argv):
            # Take out the action.
            action = self.argv.pop(0)
            if action in ['-h', '--help', '--list-commands']:
                if action == '--list-commands':
                    self.listcmds = True
                action = 'help'

            if action in ['-v', '--version']:
                try:
                    print(pkgres.get_distribution('SFLvault_common'))
                except pkgres.DistributionNotFound as e:
                    print("SFLvault-common not installed")
                
                print(pkgres.get_distribution('SFLvault_client'))

                try:
                    print(pkgres.get_distribution('SFLvault_server'))
                except pkgres.DistributionNotFound as e:
                    print("SFLvault-server not installed")
                return

            # Fix for functions
            action = action.replace('-', '_')
        # Check the first parameter, if it's in the local object.

        # Call it or show the help.
        if not hasattr(self, action):
            print("[SFLvault] Invalid command: %s" % action)
            action = 'help'

        self.action = action
        try:
            getattr(self, action)()
        except SFLvaultParserError as e:
            print("[SFLvault] Command line error: %s" % e)
            print()
            self.help(cmd=action, error=e)
        except AuthenticationError:
            raise
        except VaultError:
            #raise
            pass
        except xmlrpc.client.Fault as e:
            # On is_admin check failed, on user authentication failed.
            print("[SFLvault] XML-RPC Fault: %s" % e.faultString)
        except xmlrpc.client.ProtocolError as e:
            # Server crashed
            print("[SFLvault] XML-RPC end-point failure: %s" % e)
        except PermissionError as e:
            print("[SFLvault] Permission denied: %s" % e)
        except VaultConfigurationError as e:
            print("[SFLvault] Configuration error: %s" % e)
        except RemotingError as e:
            print("[SFLvault] Remoting error: %s" % e)
        except ServiceRequireError as e:
            print("[SFLvault] Service-chain setup error: %s" % e)
        except DecryptError as e:
            print("[SFLvault] Error decrypting messages: %s" % e.message)
        except VaultIDSpecError as e:
            print("[SFLvault] VaultID spec. error: %s" % e)
        except socket.error as e:
            print("[SFLvault] Cannot connect to the vault: %s" % e)
        except KeyringError as e:
            print("[SFLvault] Keyring error: %s" % e)
        

    def _parse(self):
        """Parse the command line options, and fill self.opts and self.args"""
        self.opts, self.args = self.parser.parse_args(args=self.argv)


    def help(self, cmd=None, error=None):
        """Print this help.

        You can use:
        
          help [command]

        to get further help for `command`."""

        # For BASH completion.
        if self.listcmds:
            # Show only a list of commands, for bash-completion.
            for x in dir(self):
                if not x.startswith('_') and callable(getattr(self, x)):
                    print(x.replace('_', '-'))
            sys.exit()

        # Normal help screen.
        print("%s version %s" % ('SFLvault-client', pkgres.get_distribution('SFLvault_client').version))
        print("---------------------------------------------")

        if not cmd:
            print("Here is a quick overview of the commands:")
            # TODO: go around all the self. attributes and display docstrings
            #       and give coherent help for every function if specified.
            #       all those not starting with _.
            for x in dir(self):
                if not x.startswith('_') and callable(getattr(self, x)):
                    doc = getattr(self, x).__doc__
                    if doc:
                        doc = doc.split("\n")[0]
                    else:
                        doc = '[n/a]'
                
                    print(" %s%s%s" % (x.replace('_','-'),
                                       (18 - len(x)) * ' ',
                                       doc))
            print("---------------------------------------------")
            print("Run: sflvault [command] --help for more details on each of those commands.")
        elif not cmd.startswith('_') and callable(getattr(self, cmd)):
            readcmd = cmd.replace('_','-')

            doc = getattr(self, cmd).__doc__
            if doc:
                print("Help for command: %s" % readcmd)
                print("---------------------------------------------")
                print(doc)
            else:
                print("No documentation available for `%s`." % readcmd)

            print("")

            try:
                self.parser.parse_args(args=['--help'])
            except ExitParserException as e:
                pass
        else:
            print("No such command")

        print("---------------------------------------------")
            
        if (error):
            print("ERROR calling %s: %s" % (cmd, error))
        return
            

    def user_add(self):
        """Add a user to the vault."""
        self.parser.set_usage("user-add [options] username")
        self.parser.add_option('-a', '--admin', dest="is_admin",
                               action="store_true", default=False,
                               help="Give admin privileges to the added user")

        self._parse()

        if (len(self.args) != 1):
            raise SFLvaultParserError("Invalid number of arguments")
        
        username = self.args[0]
        admin = self.opts.is_admin

        self.vault.user_add(username, admin)


    def customer_add(self):
        """Add a new customer."""
        self.parser.set_usage('customer-add "customer name"')
        self._parse()
        
        if (len(self.args) != 1):
            raise SFLvaultParserError('Invalid number of arguments')

        customer_name = self.args[0]

        ret = self.vault.customer_add(customer_name)
        # For the shell:
        self.next_command = "machine-add -c c#%s " % ret['customer_id']


    def user_del(self):
        """Delete an existing user."""
        self.parser.set_usage("user-del -u <username>")
        self.parser.add_option('-u', dest="username",
                               help="Username to be removed")
        self._parse()

        if not self.opts.username:
            raise SFLvaultParserError("Please specify the user with -u")

        self.vault.user_del(self.opts.username)


    def customer_del(self):
        """Delete an existing customer, it's machines and all services.

        Make sure you have detached all services' childs before removing a
        customer with machines which has services that are parents to other
        services."""
        
        self.parser.set_usage("customer-del -c <customer_id>")
        self.parser.add_option('-c', dest="customer_id", default=None,
                               help="Customer to be removed")
        self._parse()

        if not self.opts.customer_id:
            raise SFLvaultParserError("customer_id is required.")

        customer_id = self.vault.vaultId(self.opts.customer_id, 'c')
        self.vault.customer_del(customer_id)


    def machine_del(self):
        """Delete an existing machine, including all services.

        Make sure you have detached all services' childs before removing
        a machine which has services that are parents to other services.
        """        
        self.parser.set_usage("machine-del -m <machine_id>")
        self.parser.add_option('-m', dest="machine_id", default=None,
                               help="Machine to be removed")
        self._parse()

        if not self.opts.machine_id:
            raise SFLvaultParserError("machine_id is required")

        machine_id = self.vault.vaultId(self.opts.machine_id, 'm')
        self.vault.machine_del(machine_id)


    def service_del(self):
        """Delete an existing service. Make sure you have detached all
        childs before removing a parent service."""
        self.parser.set_usage("service-del -s <service_id>")
        self.parser.add_option('-s', dest="service_id", default=None,
                               help="Service to be removed")
        self._parse()

        if not self.opts.service_id:
            raise SFLvaultParserError("service_id is required")

        service_id = self.vault.vaultId(self.opts.service_id, 's')
        self.vault.service_del(service_id)
        

    def machine_add(self):
        """Add a new machine."""
        self.parser.set_usage('machine-add -n "machine name" -c <customer_id> [options]')
        self.parser.add_option('-c', '--customer', dest="customer_id",
                               help="Customer id, as 'c#123' or '123'")
        self.parser.add_option('-n', '--name', dest="name",
                               help="Machine name, used for display everywhere")
        self.parser.add_option('-d', '--fqdn', dest="fqdn", default='',
                               help="Fully qualified domain name, if available")
        self.parser.add_option('-i', '--ip', dest="ip", default='',
                               help="Machine's IP address, in order to access itfrom it's hierarchical position")
        self.parser.add_option('-l', '--location', dest="location", default='',
                               help="Machine's physical location, position in racks, address, etc..")
        self.parser.add_option('--notes', dest="notes",
                               help="Notes about the machine, references, URLs.")
        self._parse()

        if not self.opts.name:
            raise SFLvaultParserError("Required parameter 'name' omitted")
        
        ## TODO: make a list-customers and provide a selection using arrows or
        #        or something alike.
        if not self.opts.customer_id:
            raise SFLvaultParserError("Required parameter 'customer' omitted")

        o = self.opts
        customer_id = self.vault.vaultId(o.customer_id, 'c')
        ret = self.vault.machine_add(customer_id, o.name, o.fqdn,
                                     o.ip, o.location, o.notes)
        # For the shell:
        self.next_command = "service-add -m m#%s -u " % ret['machine_id']



    def _service_clean_url(self, url):
        """Remove password in URL, and notify about rewrite."""
        # Rewrite url if a password was included... strip the port and
        #       username from the URL too.
        url = urllib.parse.urlparse(url)
        if url.password:
            out = []
            if url.username:
                out.append('%s@' % url.username)
            
            out.append(url.hostname)
            
            if url.port:
                out.append(":%d" % url.port)
            
            url = urllib.parse.urlunparse((url[0], ''.join(out),
                                           url[2], url[3], url[4], url[5]))

            print("NOTE: Do not specify password in URL. Rewritten as: %s" % url)
            return url
        else:
            return False


    def service_add(self):
        """Add a service to a particular machine.

        The secret/password/authentication key will be asked in the
        interactive prompt.

        Note : Passwords will never be taken from the URL when invoked on the
               command-line, to prevent sensitive information being held in
               history.
        """
        self.parser.set_usage("service-add -u <url> -m <machine_id> -g <group_id> [options]")
        self.parser.add_option('-m', '--machine', dest="machine_id",
                               help="Attach Service to Machine #, as "\
                                    "'m#123', '123' or an alias")
        self.parser.add_option('-u', '--url', dest="url",
                               help="Service URL, full proto://[username@]"\
                               "fqdn.example.org[:port][/path[#fragment]], "\
                               "WITHOUT the secret.")

        self.parser.add_option('-p', '--parent', dest="parent_id",
                               help="Make this Service child of Parent "\
                                    "Service #")
        self.parser.add_option('-g', '--group', dest="group_ids",
                               action="append", type="string",
                               help="Access group_id for this service, as "\
                               "'g#123' or '123'. Use group-list to view "\
                               "complete list. You can specify multiple groups")
        self.parser.add_option('--notes', dest="notes",
                               help="Notes about the service, references, "\
                                    "URLs.")
        self.parser.add_option('--metadata', dest="metadata",
                               action="append", type="string",
                               help="Add metadata to this service. Specify as key=value. "\
                                    "This option can appear more than once.")
        self._parse()

        for x in ('url', 'machine_id', 'group_ids'):
            if not getattr(self.opts, x):
                raise SFLvaultParserError("Required parameter '%s' omitted" % x)
        o = self.opts

        rewritten_url = self._service_clean_url(o.url)
        url = rewritten_url if rewritten_url else o.url

        secret = ask_for_service_password(edit=True, url=url)

        machine_id = 0
        parent_id = 0
        group_ids = []
        if o.machine_id:
            machine_id = self.vault.vaultId(o.machine_id, 'm')
        if o.parent_id:
            parent_id = self.vault.vaultId(o.parent_id, 's')
        if o.group_ids:
            group_ids = [self.vault.vaultId(g, 'g') for g in o.group_ids]
        metadata = {}
        if o.metadata:
            for s in o.metadata:
                if '=' not in s:
                    raise SFLvaultParserError("Missing '=' in " \
                                              "metadata option: '%s'" % s)
                key, val = s.split('=', 1)
                metadata[key] = val

        # WARNING: we should check if server supports "metadata" before sending
        self.vault.service_add(machine_id, parent_id, url, group_ids, secret,
                               o.notes, metadata)

    def service_edit(self):
        """Edit service informations."""
        # TODO: implement -s, -m an d-c for those EDIT things..
        self._something_edit("service-edit [service_id]",
                             'service_id', 's',
                             self.vault.service_get,
                             self.vault.service_put,
                             ui.ServiceEditDialogDisplay,
                             'service-edit aborted',
                             decrypt=False)
    def machine_edit(self):
        """Edit machine informations."""
        self._something_edit("machine-edit [machine_id]",
                             'machine_id', 'm',
                             self.vault.machine_get,
                             self.vault.machine_put,
                             ui.MachineEditDialogDisplay,
                             'machine-edit aborted')
    def customer_edit(self):
        """Edit customer informations."""
        self._something_edit("customer-edit [customer_id]",
                             'customer_id', 'c',
                             self.vault.customer_get,
                             self.vault.customer_put,
                             ui.CustomerEditDialogDisplay,
                             'customer-edit aborted')
    def group_edit(self):
        """Edit Group informations"""
        self._something_edit("group-edit [group_id]",
                             'group_id', 'g',
                             self.vault.group_get,
                             self.vault.group_put,
                             ui.GroupEditDialogDisplay,
                             'group-edit aborted')
    def _something_edit(self, usage, required_args, vault_id_type,
                        get_function, put_function, ui_class, abort_message,
                        **kwargs):

        self.parser.set_usage(usage)
        self._parse()

        if not len(self.args):
            raise SFLvaultParserError("Required argument: %s" % required_args)

        thing_id = self.vault.vaultId(self.args[0], vault_id_type)

        # kwargs can contain 'decrypt=False', see service_edit.
        thing = get_function(thing_id, **kwargs)

        if get_function == self.vault.service_get and \
                not thing.get('cryptsymkey'):
            raise PermissionError("You don't have access to that service.  You cannot modify it.")

        try:
            dialog = ui_class(thing)
            save, data = dialog.run()
        except TypeError as e:
            print("W00ps!  The shell GUI doesn't support certain operations, like copy and pasting, or specially weird characters.")
            print("In fact, most terminal support doing SHIFT+Middle_Button to paste the text as if the user has typed it, instead of sending Button2 events.")
            print("Try it out, hope this helps.")
            return

        if save:
            print("Sending data to vault...")
            put_function(thing_id, data)
        else:
            print(abort_message)


    def service_passwd(self):
        """Change the password for a service.

        Do not specify password on command line, it will be asked on the
        next line.
        """
        self.parser.set_usage("service-passwd <service_id>")
        self._parse()

        if not len(self.args):
            raise SFLvaultParserError("Required argument: service_id")

        service_id = self.vault.vaultId(self.args[0], 's')

        serv = self.vault.service_get(service_id)

        newsecret = ask_for_service_password(edit=True, url=serv['url'])

        self.vault.service_passwd(service_id, newsecret)


    def alias(self):
        """Set an alias, local shortcut to VaultIDs (s#123, m#87, etc..)

        List, view or set an alias."""
        self.parser.set_usage("alias [options] [alias [VaultID]]")

        self.parser.add_option('-d', '--delete', dest="delete",
                               metavar="ALIAS", help="Delete the given alias")

        self._parse()

        if self.opts.delete:
            res = self.vault.cfg.alias_del(self.opts.delete)

            if res:
                print("Alias removed")
            else:
                print("No such alias")

        elif len(self.args) == 0:
            # List aliases
            l = self.vault.cfg.alias_list()
            print("Aliased VaultIDs:")
            for x in l:
                print("\t%s\t%s" % (x[0], x[1]))

        elif len(self.args) == 1:
            # Show this alias's value
            a = self.vault.cfg.alias_get(self.args[0])
            if a:
                print("Aliased VaultID:")
                print("\t%s\t%s" % (self.args[0], a))
            else:
                print("Invalid alias")

        elif len(self.args) == 2:
            try:
                r = self.vault.cfg.alias_add(self.args[0], self.args[1])
            except ValueError as e:
                raise SFLvaultParserError(str(e))

            print("Alias added")

        else:
            raise SFLvaultParserError("Invalid number of parameters")


    def customer_list(self):
        """List existing customers.

        This option takes no argument, it just lists customers with their IDs."""
        self._parse()
        
        if len(self.args):
            raise SFLvaultParserError('Invalid number of arguments')

        self.vault.customer_list()

    def user_list(self):
        """List existing users.

        This option takes no argument, it lists the current users and their
        privileges."""
        self.parser.set_usage("user-list [-g]")
        self.parser.add_option('-g', '--groups', default=False,
                               action="store_true", dest="groups",
                               help="List user's group infos")
        self._parse()

        if len(self.args):
            raise SFLvaultParserError("Invalid number of arguments")

        self.vault.user_list(self.opts.groups)


    def _group_service_options(self):
        self.parser.add_option('-g', dest="group_id",
                               help="Group to add the service to")
        self.parser.add_option('-s', dest="service_id",
                               help="Service to be added")

    def _group_service_parse(self):
        if not self.opts.group_id or not self.opts.service_id:
            raise SFLvaultParserError("-g and -s options required")

        self.opts.group_id = self.vault.vaultId(self.opts.group_id, 'g')
        self.opts.service_id = self.vault.vaultId(self.opts.service_id, 's')

    def group_add_service(self):
        """Add a service to a group, doing necessary re-encryption"""
        self.parser.set_usage("group-add-service -g <group_id> -s <service_id>")
        self._group_service_options()
        self._parse()

        self._group_service_parse()
        
        self.vault.group_add_service(self.opts.group_id, self.opts.service_id)

    def group_del_service(self):
        """Remove a service from a group"""
        self.parser.set_usage("group-del-service -g <group_id> -s <service_id>")
        self._group_service_options()
        self._parse()

        self._group_service_parse()
        
        self.vault.group_del_service(self.opts.group_id, self.opts.service_id)


    def _group_user_options(self):
        self.parser.add_option('-g', dest="group_id",
                               help="Group to add the service to")
        self.parser.add_option('-u', dest="user",
                               help="User to be added")

    def _group_user_parse(self):
        if not self.opts.group_id or not self.opts.user:
            raise SFLvaultParserError("-g and -u options required")
        
        self.opts.group_id = self.vault.vaultId(self.opts.group_id, 'g')

    def group_add_user(self):
        """Add a user to a group, doing necessary re-encryption"""
        self.parser.set_usage("group-add-user [-a] -g <group_id> -u <user>")
        self.parser.add_option('-a', action="store_true", dest='is_admin',
                               default=False, help="Mark as group admin")
        self._group_user_options()
        self._parse()
        self._group_user_parse()
        
        self.vault.group_add_user(self.opts.group_id, self.opts.user,
                                  self.opts.is_admin)

    def group_del_user(self):
        """Remove a user from a group"""
        self.parser.set_usage("group-del-user -g <group_id> -u <user>")
        self._group_user_options()
        self._parse()

        self._group_user_parse()
        
        self.vault.group_del_user(self.opts.group_id, self.opts.user)

    def group_del(self):
        """Remove a group from the vault

        For this to be successful, the group must have no more services
        associated with it."""
        self.parser.set_usage("group-del -g <group_id>")
        self.parser.add_option('-g', dest="group_id", default=None,
                               help="Group to be removed")
        self._parse()
        
        if not self.opts.group_id:
            raise SFLvaultParserError("group_id is required")

        group_id = self.vault.vaultId(self.opts.group_id, 'g')
        self.vault.group_del(group_id)

    def group_add(self):
        """Add a group to the vault

        This command accepts a group name (as string) as first and only
        parameter.
        """
        self.parser.set_usage('group-add "group name"')
        self._parse()

        if len(self.args) != 1:
            raise SFLvaultParserError("Group name (as string) required")

        self.vault.group_add(self.args[0])


    def group_list(self):
        """List existing groups."""
        self.parser.set_usage("group-list [options]")
        self.parser.add_option('-q', dest='quiet', default=False,
                               action="store_true", help="Hide members")
        self._parse()

        if len(self.args):
            raise SFLvaultParserError("Invalid number of arguments")

        self.vault.group_list(self.opts.quiet)


    def machine_list(self):
        """List existing machines.

        This command will list all machines in the vault's database."""
        ## TODO: add support for listing only machines of a certain c#id
        #        (customer_id)
        self.parser.set_usage("machine-list [options]")
        self.parser.add_option('-v', '--verbose', action="store_true",
                               dest='verbose', default=False,
                               help="Enable verbose output (location and notes)")
        self.parser.add_option('-c', '--customer', dest='customer_id',
                               help="Customer id")
        self._parse()

        customer_id = None
        if self.opts.customer_id:
            customer_id = self.vault.vaultId(self.opts.customer_id, 'c')

        if len(self.args):
            raise SFLvaultParserError("Invalid number of arguments")

        self.vault.machine_list(self.opts.verbose, customer_id)


    def user_passwd(self):
        """Change the passphrase protecting your local private key"""
        self.parser.set_usage("user-passwd")
        self._parse()

        if len(self.args) != 0:
            raise SFLvaultParserError("user-passwd takes no arguments")

        self.vault.user_passwd()

    def user_setup(self):
        """Setup a new user on the vault.

        Call this after an admin has called `user-add` on the vault.
        
        Arguments:
            username: the username used in the `user-add` call.
        
            vault_url: the URL (http://example.org:port/vault/rpc) to the vault.
        """
        
        self.parser.set_usage("user-setup <username> <vault_url>")
        self._parse()
        
        if len(self.args) != 2:
            raise SFLvaultParserError("Invalid number of arguments")

        username = self.args[0]
        url      = self.args[1]

        self.vault.user_setup(username, url)

    def show(self):
        """Show informations to connect to a particular service.

        VaultID - service ID as 's#123', '123', or alias pointing to a service
                  ID."""
        self.parser.set_usage("show [options] <service_id>")
        self.parser.add_option('-q', '--quiet', dest="quiet",
                               action="store_false", default=True,
                               help="Show notes, locations, groups")
        self._parse()

        if len(self.args) != 1:
            raise SFLvaultParserError("Invalid number of arguments")

        vid = self.vault.vaultId(self.args[0], 's')

        self.vault.show(vid, self.opts.quiet, self.opts.quiet)




    def connect(self):
        """Connect to a remote SSH host, sending password on the way.

        VaultID - service ID as 's#123', '123', or alias pointing to a service
                  ID."""
        # Chop in two parts
        self.parser.set_usage("connect [opts] VaultID")
        self.parser.add_option('-a', '--alias', dest="alias",
                               default=None,
                               help="Add an alias to that service at the same "\
                                    "time")
        self.parser.add_option('-s', '--show', dest="show",
                               default=False, action="store_true",
                               help="Show the credentials before connecting")
        self._parse()

        if len(self.args) < 1:
            raise SFLvaultParserError("Invalid number of arguments")
        
        where = self.args[0]
        vid = self.vault.vaultId(where, 's') # In case we want to add an alias
        command_line = self.args[1:]

        if self.opts.alias:
            try:
                al = "s#%d" % vid
                if command_line:
                    al += " " + ' '.join(command_line)
                r = self.vault.cfg.alias_add(self.opts.alias, al)
            except ValueError as e:
                raise SFLvaultParserError(str(e))
            print("Alias added")

        # Grab alias's arguments if they exist...
        alias = self.vault.cfg.alias_get(where)
        if alias:
            chunks = alias.split()
            if len(chunks) > 1:
                command_line += chunks[1:]
            where = chunks[0]

        vid = self.vault.vaultId(where, 's')

        self.vault.connect(vid, self.opts.show, command_line=command_line)



    def search(self):
        """Search the vault for the specified keywords."""
        self.parser.set_usage('search [opts] keyword1 ["key word2" ...]')
        self.parser.add_option('-g', '--group', dest="groups",
                               action="append", type="string",
                               help="Search in these groups only")
        self.parser.add_option('-q', '--quiet', dest="verbose",
                               action="store_false", default=True,
                               help="Don't show verbose output (includes notes, location)")
        
        self.parser.add_option('-m', '--machine', dest="machines",
                               action="append", type="string",
                               help="Filter results on these machines only")

        self.parser.add_option('-c', '--customer', dest="customers",
                               action="append", type="string",
                               help="Filter results on these customers only")

        self._parse()

        if not len(self.args):
            raise SFLvaultParserError("Search terms required")

        # Get the values for each filter spec..
        fields = {'groups': 'g',
                  'machines': 'm',
                  'customers': 'c'}
        filters = {}
        for f in list(fields.keys()):
            criteria = None
            if getattr(self.opts, f):
                criteria = [self.vault.vaultId(x, fields[f])
                            for x in getattr(self.opts, f)]
            filters[f] = criteria

        self.vault.search(self.args, filters or None, self.opts.verbose)

    def wallet(self):
        """Put your SFLvault password in a wallet"""
        self.parser.set_usage('wallet [num]')
        self._parse()

        if len(self.args) == 0:
            for i, name, obj, status, current in self.vault.cfg.wallet_list():
                print("%s. %s - %s%s" % (i, name, status,
                                         " (*current)" if current else ''))
            print()
            print("Execute 'wallet <num>' to change wallet")
            return

        # Otherwise
        id = self.args[0]
        lst = self.vault.cfg.wallet_list()
        ids = [x[0] for x in lst]

        if id == '0':
            self.vault.cfg.wallet_set(None, None)
            self.vault.set_getpassfunc(None)
            print("Keyring disabled")
            return

        if id not in ids:
            raise SFLvaultParserError("Invalid wallet ID. Try `wallet` without parameters")

        # Make sure the wallet is supported
        self.vault.cfg.wallet_test([x[1] for x in lst if x[0] == id][0])

        try:
            passwd = getpass.getpass("Enter your current vault password: ")
            self.vault.cfg.wallet_set(id, passwd)
        except KeyringError as e:
            print("[SFLvault] %s" % e)
            return False
        print("Keyring set and password saved")



class SFLvaultCompleter:
    def __init__(self, namespace):
        self.namespace = namespace
        self.matches = []

    def complete(self, text, state):
        if state == 0:
            self.matches = self.global_matches(text)
        try:
            return self.matches[state]
        except IndexError:
            return None

    def global_matches(self, text):
        matches = []
        for word in self.namespace:
            if word.find(text,0,len(text)) == 0:
                matches.append(word)
        return matches

###
### Execute requested command-line command
###    

# Default configuration file
if platform.system() == 'Windows':
    CONFIG_FILE = '~/Application Data/SFLvault/config.ini'
else:
    CONFIG_FILE = '~/.sflvault/config'
# Environment variable to override default config file.
CONFIG_FILE_ENV = 'SFLVAULT_CONFIG'

def main():
    # Call the appropriate function of the 'f' object, according to 'action'
    func_list = []
    for onefunc in dir(SFLvaultCommand):
        if onefunc[0] != '_':
            func_list.append(onefunc.replace('_', '-'))

    if platform.system() != 'Windows':
        readline.set_completer_delims('_')
        readline.set_completer(SFLvaultCompleter(func_list).complete)
        readline.parse_and_bind("tab: complete")

    # Set the output to UTF-8 if it's not set by the terminal (for PIPES
    # redirection)
    # See http://stackoverflow.com/questions/492483/setting-the-correct-encoding-when-piping-stdout-in-python
    if sys.stdout.encoding is None:
        import codecs
        sys.stdout = codecs.getwriter('utf8')(sys.stdout)

    # Extract the '-i "identity"' before starting if present.
    args = sys.argv[:]
    identity = None
    if len(args) > 1 and args[1] == '-i':
        del args[1]
        if len(args) == 1 or args[1].startswith('-'):
            print("Error: Identity required after -i")
            sys.exit()
        identity = args.pop(1)

    config_file = CONFIG_FILE
    if identity:
        config_file = "%s.%s" % (CONFIG_FILE, identity)
        print("NOTICE: USING VAULT IDENTITY: %s  (in %s)" % (identity, config_file))
    elif CONFIG_FILE_ENV in os.environ:
        config_file = os.environ[CONFIG_FILE_ENV]

    if len(args) == 1 or args[1] == 'shell':
        s = SFLvaultShell(config=config_file)
        try:
            s._run()
        except (KeyboardInterrupt, EOFError) as e:
            print("\nExiting.")
            sys.exit()
    else:
        f = SFLvaultCommand(config=config_file)
        f._run(args[1:])
    

# For wrappers.
if __name__ == "__main__":

    main()



# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Fallback Shell is used when we back off an active shell with the
escape sequence, to get a grip on the built-in FISH functionalities.

Other commands will be implemented in the future, giving ability to
script on the client side things that will be sent to through the shell."""

from sflvault.client.fishlib import FishClient, showstatus
from sflvault.client.commands import SFLvaultCompleter, NoExitParser
from sflvault.client.commands import ExitParserException, SFLvaultParserError
from sflvault.client.utils import ServiceSwitchException
import shlex
import sys
import optparse
import os
import platform
if platform.system() != 'Windows':
    import readline

class SFLvaultFallback(object):

    def __init__(self, chain, shell_obj):
        # we're to replace completer with new one
        self.parent_completer = readline.get_completer()
        self.parent_completer_last_item = readline.get_current_history_length()\
                                          - 1
        self.func_list = []
        # Services chain
        self.chain = chain
        self.current_shell = shell_obj

        for onefunc in dir(SFLvaultFallback):
            if onefunc[0] != '_':
                self.func_list.append(onefunc)

        readline.set_completer(SFLvaultCompleter(self.func_list).complete)

        self.fish_obj = FishClient(shell_obj)
        self.fish_obj.start()


    # TODO: we need an __getattr__ thing, that's going to catch
    # functions that aren't already defined here.

    def _parse(self):
        """Parse the command line options, and fill self.opts and self.args"""
        (self.opts, self.args) = self.parser.parse_args(args=self.argv)

    def _return_completer(self):
        last_item = readline.get_current_history_length() - 1
        if last_item > self.parent_completer_last_item:
            for tmpid in range(self.parent_completer_last_item, last_item):
                try:
                    readline.remove_history_item(readline.get_current_history_length() - 1)
                except ValueError as e:
                    pass
        readline.set_completer(self.parent_completer)
        print("\nExiting SFLvaultFallback.")

    def _run(self):
        """Run the shell, and dispatch commands"""

        while True:
            print("\nWelcome to SFLvaultFallback. Type 'help' for help.")
            prompt = "SFLvaultFallback> "

            while True:
                cmd = input(prompt)
                if not cmd:
                    continue

                # Get sys.argv-like parameters
                args = shlex.split(cmd)

                # Local (shell) cmds take precedence over SFLvaultCommand cmds.
                if args[0] == 'quit':
                    self._return_completer()
                    return

                if args[0] in self.func_list:
                    self.parser = NoExitParser(usage=optparse.SUPPRESS_USAGE)
                    self.argv = args
                    self.args = []
                    self.opts = object()

                    try:
                        getattr(self, args[0])()
                    except SFLvaultParserError as e:
                        self.help(cmd=args[0], error=e)
                        pass
                    except ExitParserException as e:
                        pass
                    except UnboundLocalError as e:
                        # TODO: verify if this error can be worked out
                        # differently
                        self.help(cmd=args[0])

                else:
                    self.help()

    def help(self, cmd=None, error=None):
        """Display the list of available commands"""
        
        if not cmd:
            for func in self.func_list:
                doc = getattr(self, func).__doc__
                if doc:
                    doc = doc.split("\n")[0]
                else:
                    doc = '[n/a]'
            
                print("  %s%s%s" % (func,(25 - len(func)) * ' ',doc))
        elif not cmd.startswith('_') and callable(getattr(self, cmd)):
            doc = getattr(self, cmd).__doc__
            if doc:
                print("Help for command: %s" % cmd)
                print("---------------------------------------------")
                print(doc)
            else:
                print("No documentation available for `%s`." % cmd)

            print("")
            try:
                self.parser.parse_args(args=['--help'])
            except ExitParserException as e:
                pass
            
        if (error):
            print("ERROR calling %s: %s" % (cmd, error))


    def shells(self):
        """List available shells and switch between them"""
        last = None
        current = self.current_shell
        star = False
        count = 0
        out = {}
        for srv in reversed(self.chain.service_list):
            if hasattr(srv, 'shell_handle'):
                star = (current == srv.shell_handle)
                if last == srv.shell_handle:
                    continue
                count += 1
                print("%d. Shell on service %s%s" % (count, srv,
                                                   ' (current)' if star else ''))
                last = srv.shell_handle
                out[str(count)] = srv
        while True:
            shell = input("Switch> ")
            if shell in out:
                srv = out[shell]
                e = ServiceSwitchException("Switching to service %s" % srv)
                e.service = srv
                raise e
            print("Please select a shell from the list")
            break

    def quit(self):
        # Added only for auto completition
        """Quit SFLvaultFallback shell and get back to SSH"""
        pass

    def get(self):
        """Downloading files via SSH using FISH protocol"""
        self.parser.set_usage("get [options]")
        self.parser.add_option('-s', '--source', dest="source",
                               help="Remote machine source filename")
        self.parser.add_option('-d', '--dest', dest="dest",
                               help="Local machine destination filename")

        self._parse()
        
        if len(self.args) != 1:
            raise SFLvaultParserError("Invalid number of arguments")

        if not self.opts.source:
            raise SFLvaultParserError("Required parameter 'source' omitted")

        if not self.opts.dest:
            # we will store in /tmp folder by default
            tmp = self.opts.source.split('/')
            self.opts.dest = '/tmp/' + tmp[-1]
            print("Source file will be stored as: %s" % self.opts.dest)
            #raise SFLvaultParserError("Required parameter 'dest' omitted")
        
        try:
            local_file = open(self.opts.dest, 'w')
        except IOError as e:
            raise SFLvaultParserError("I can't create file: %s" % \
                                      self.opts.dest)
        try:
            self.fish_obj.retr(self.opts.source, local_file, showstatus)
        except ValueError as e:
            raise SFLvaultParserError("I can't find file: %s" % \
                                      self.opts.source)

        local_file.close()

    def put(self):
        """Uploading files via SSH using FISH protocol"""
        
        self.parser.set_usage("put [options]")
        self.parser.add_option('-s', '--source', dest="source",
                               help="Locale machine source filename")
        self.parser.add_option('-d', '--dest', dest="dest",
                               help="Remote machine destination filename")

        self._parse()
        
        if len(self.args) != 1:
            raise SFLvaultParserError("Invalid number of arguments")

        if not self.opts.source:
            raise SFLvaultParserError("Required parameter 'source' omitted")

        if not self.opts.dest:
            # we will store in /tmp folder by default
            tmp = self.opts.source.split('/')
            self.opts.dest = '/tmp/' + tmp[-1]
            print("Source file will be stored as: %s" % self.opts.dest)
            #raise SFLvaultParserError("Required parameter 'dest' omitted")
        
        try:
            filesize = os.path.getsize(self.opts.source)
        except OSError as e:
            raise SFLvaultParserError("I can't read file: %s" % self.opts.source)

        local_file = open(self.opts.source, 'rb')
        self.fish_obj.stor(self.opts.dest, local_file, filesize, showstatus)
        local_file.close()




#!/usr/bin/python
# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008-2009  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
if platform.system() != 'Windows':
    import pexpect
import os
from pprint import pprint
import pdb
from datetime import datetime
#import subprocess


class FishClient(object):
    """Fish protocol implementation, given a running SSH session
    in an pexpect spawn'd object.
    """
    
    def __init__(self, proc):
        """Init the class with the process. Start the session with .start()"""
        self.proc = proc

    def _wait_for(self, until):
        """Read lines until one starting with 'until', return read lines.
        """
        lines = []
        while True:
            # Add some handling to make it stop at some point..
            try:
                l = self.proc.readline()
            except KeyboardInterrupt as e:
                print("Before: ", self.proc.before)
                print("After: ", self.proc.after)
                print("Buffer: ", self.proc.buffer)
                return (lines, 'KeyboardInterrupt')
            # Stop when starts with 'until'
            if l.startswith(until):
                return (lines, l)
            lines.append(l)


    def start(self):
        """Start the FISH connection, init and get the remote version if any
        """
        self.proc.sendline('#FISH')
        self.proc.sendline('echo; start_fish_server; echo "### 200"')
        lines, retval = self._wait_for('### 200')

        self.proc.sendline('#VER 0.0.2 <feature1> <feature2> <...>')
        self.proc.sendline("echo '### 000'")
        lines, retval = self._wait_for('###')

    def dele(self, filename):
        """Delete a remote file"""
        self.proc.sendline("#DELE %s" % filename)
        self.proc.sendline("rm -f %s 2>/dev/null" % filename)
        self.proc.sendline("echo '### 000'")
        lines, retval = self._wait_for('### 000')
        pprint((lines, retval))


    def retr(self, filename, write_to, callback=None):
        """Retrieve a file from remote host, and write the content to the
        given file-like object"""
        self.proc.sendline('''#RETR "%s"''' % (filename))
        self.proc.sendline("""ls -l "%s" | ( read a b c d x e; echo $x ); echo '### 100'; cat "%s"; echo '### 200'""" % (filename, filename))
        lines, retval = self._wait_for('### 100')
        filelen = int(lines.pop().strip())
        print(filelen)

        print("Downloading...")
        dl = 0
        remain = filelen
        self.callback_chk = -1
        while True:
            if callback:
                self.remain = remain
                self.filelen = filelen
                callback(self)

            # Less than one byte remaining (should always be zero btw)
            if remain < 1:
                break
            packsize = (4096 if remain > 4095 else remain % 4096)
            chunk = self.proc.read_nonblocking(packsize, timeout=None)
            # ttys always send \r\n even though the underlying program
            # sends a \n, that's why this mangling is necessary.
            chunk = chunk.replace("\r\n", "\n")
            remain -= len(chunk)
            # This if is necessary to catch \r\n splitted into two read() calls
            if chunk[-1:] == "\r":
                nextchr = self.proc.read_nonblocking(1, timeout=None)
                if nextchr == "\n":
                    chunk = chunk[:-1] + "\n"
                elif remain == 0:
                    self.proc.buffer = nextchr
                else:
                    remain -= 1
                    chunk += nextchr
            write_to.write(chunk)

        print("Download done.")

        # Take out the CRLF chars produced by cat.
        #self.proc.read(2)

        lines, retval = self._wait_for('### 200')

    def stor(self, remote_filename, read_from, filelen, callback=None):
        """Store a file to remote host, reading from read_form.read()"""
        olddelay = self.proc.delaybeforesend

        # Taken from lftp's fish.c implementation
              
        cmd = """#STOR %lu "%s"\n""" % (filelen, remote_filename)
        cmd += """stty -echo; rest=%lu;file="%s";:>$file;echo '### 001';""" % (filelen,remote_filename)
        cmd += "stty -icanon;"
        cmd += "if echo 1|head -c 1 -q ->/dev/null 2>&1;then "
        cmd += "head -c $rest -q -|(cat>$file;cat>/dev/null);"
        cmd += "else while [ $rest -gt 0 ];do "
        cmd += "bs=4096;cnt=`expr $rest / $bs`;"
        cmd += "[ $cnt -eq 0 ] && { cnt=1;bs=$rest; }; "
        cmd += "n=`dd ibs=$bs count=$cnt 2>/dev/null|tee -a $file|wc -c`;"
        cmd += "[ \"$n\" -le 0 ] && exit;"
        cmd += "rest=`expr $rest - $n`; "
        cmd += "done;fi;stty icanon;stty echo;echo '### 200'\n"

        cmd2 = """#STOR %lu "%s"
              echo '### 001'
              {
                     stty -echo
                     file="%s"
                     rest=%lu
                     while [ $rest -gt 0 ]
                     do
                         cnt=`expr \\( $rest + 255 \\) / 256`
                         n=`dd bs=256 count=$cnt 2> /dev/null | tee -a "$file" | wc -c`
                         rest=`expr $rest - $n`
                     done
                     stty echo
              }
              echo '### 200'\n""" % (filelen, remote_filename,
                                        remote_filename, filelen)


        self.proc.send(cmd)
        lines, retval = self._wait_for('### 001')

        # Write:
        print("Uploading...")
        self.proc.delaybeforesend = 0
        dl = 0
        remain = filelen
        self.callback_chk = -1
        while True:
            if callback:
                self.remain = remain
                self.filelen = filelen
                callback(self)

            # Less than one byte remaining (should always be zero btw)
            if remain < 1:
                break
            packsize = (4096 if remain > 4095 else remain % 4096)
            chunk = read_from.read(packsize)
            remain -= len(chunk)
            os.write(self.proc.child_fd, chunk)

        self.proc.delaybeforesend = olddelay
        lines, retval = self._wait_for('### 200')
        print("Uploading done.")

def showstatus(fishproc):
    n = datetime.now().second
    if fishproc.callback_chk != n:
        fishproc.callback_chk = n
        diff = fishproc.filelen - fishproc.remain
        print("%d / %d" % (diff, fishproc.filelen))

if __name__ == '__main__':
    # Connect to an ssh command
    proc = pexpect.spawn('ssh office', encoding='utf-8')

    # Create the FishClient and wrap it around.
    # This can also be done after the interact() session if you wish.
    fc = FishClient(proc)
    
    # Set the escape character to what you want.
    proc.interact(escape_character=chr(30))
    
    # Automatically start the Fish session when escaping.
    fc.start()

    oldn = -1

    fname = '/tmp/' + input("Fichier dans /tmp/: ")
    filesize = os.path.getsize(fname)
    myfile = open(fname, 'rb')
    fc.stor(fname, myfile, filesize, showstatus)
    myfile.close()

    proc.interact()

    print("Closing connection")
    proc.close()




# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008-2009  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pkg_resources as pkgres
from configparser import ConfigParser, NoSectionError
import xmlrpc.client
import getpass
import sys
import re
import os
import time

from subprocess import Popen, PIPE

from decorator import decorator
from pprint import pprint

from sflvault.common import VaultError
from sflvault.common.crypto import *
from sflvault.client.utils import *
from sflvault.client import remoting





### Setup variables and functions


def vaultReply(rep, errmsg="Error"):
    """Tracks the Vault reply, and raise an Exception on error"""

    if rep['error']:
        print("%s: %s" % (errmsg, rep['message']))
        raise VaultError(rep['message'])
    
    return rep


#
# authenticate decorator
#
def authenticate(keep_privkey=False):
    def do_authenticate(func, self, *args, **kwargs):
        """Login decorator
        
        self is there because it's called on class elements.
        """

        username = self.cfg.get('SFLvault', 'username')
        privkey = None

        # Check if we've cached the decrypted private key
        if hasattr(self, 'privkey'):
            # Use cached private key.
            privkey = self.privkey

        else:

            try:
                privkey_enc = self.cfg.get('SFLvault', 'key')
            except:
                raise VaultConfigurationError("No private key in local config, init with: user-setup username vault-url")
        
            try:
                privpass = self.getpassfunc()
                privkey_packed = decrypt_privkey(privkey_enc, privpass)
                del(privpass)
                eg = ElGamal.ElGamalobj()
                (eg.p, eg.x, eg.g, eg.y) = unserial_elgamal_privkey(privkey_packed)
                privkey = eg

            except DecryptError as e:
                print("[SFLvault] Invalid passphrase")
                return False
            except TypeError as e:
                print("[SFLvault] Could not retrieve passphrase, please resync your Keyring using the 'wallet' command.")
                return False
            except KeyboardInterrupt as e:
                print("[aborted]")
                return False

            # When we ask to keep the privkey, keep the ElGamal obj.
            if keep_privkey or self.shell_mode:
                self.privkey = privkey

       # Go for the login/authenticate roundtrip

        # TODO: check also is the privkey (ElGamal obj) has been cached
        #       in self.privkey (when invoked with keep_privkey)
        retval = self.vault.login(username, pkgres.get_distribution('SFLvault_client').version)
        self.authret = retval
        if not retval['error']:
            # try the last token
            retval2 = self.vault.authenticate(username, self.authtok)
            if retval2['error']:
                print(retval2['message'])
                # decrypt token.
                cryptok = privkey.decrypt(unserial_elgamal_msg(retval['cryptok']))
                retval3 = self.vault.authenticate(username, b64encode(cryptok))
                self.authret = retval3

                if retval3['error']:
                    #print retval
                    print(retval3)
                    raise AuthenticationError("Authentication failed: %s" % \
                                              retval3['message'])
                else:
                    self.authtok = retval3['authtok']
                    print(retval3['message'])

            else:
                self.authtok = retval2['authtok']
                print(retval2['message'])
        else:
            raise AuthenticationError("Authentication failed: %s" % \
                                      retval['message'])

        return func(self, *args, **kwargs)

    return decorator(do_authenticate)

###
### Différentes façons d'obtenir la passphrase
###
class AskPassMethods(object):
    """Wrapper for askpass methods"""
    
    env_var = 'SFLVAULT_ASKPASS'

    def program(self):
        try:
            p = Popen(args=[self._program_value], shell=False, stdout=PIPE)
            p.wait()
            return p.stdout.read()
        except OSError as e:
            msg = "Failed to run '%s' : %s" % (os.environ[self.env_var], e)
            raise ValueError(msg)

    def default(self):
        """Default function to get passphrase from user, for authentication."""
        return getpass.getpass("Vault passphrase: ", stream=sys.stderr)
    
    def __init__(self, config):
        # Default
        self.getpass = self.default
        self.cfg = config

        # Use 'program' is SFLVAULT_ASKPASS env var exists
        env_var = AskPassMethods.env_var
        if env_var in os.environ:
            self._program_value = os.environ[env_var]
            self.getpass = self.program
            return

        wallet_name = self.cfg.wallet_get()

        def keyring_wallet():
            self.cfg._check_keyring()
            import keyring
            backend = self.cfg.wallet_get_obj()
            return (backend.get_password("sflvault", self.cfg._wallet_key)
                    or backend.get_password("sflvault",
                                            self.cfg._wallet_key_legacy))

        if wallet_name:
            self.getpass = keyring_wallet

###
### Configuration manager for SFLvault
###
class SFLvaultConfig(object):
    """This object deals with everything configuration.

    It handles the aliases adding/removal and talks with other systems
    through Exceptions.

    It handles the keyring support and passwords management

    """
    def __init__(self, config_file):
        self.config_file = os.path.expanduser(config_file)
        self.config_check()
        self.config_read()
        
        
    def config_check(self):
        """Checks for ownership and modes for all paths and files, à-la SSH"""
        fullfile = self.config_file
        fullpath = os.path.dirname(self.config_file)
    
        if not os.path.exists(fullpath):
            os.makedirs(fullpath, mode=0o700)

        if not os.stat(fullpath)[0] & 0o700:
            ### TODO: RAISE EXCEPTION INSTEAD
            print("Modes for %s must be 0700 (-rwx------)" % fullpath)
            sys.exit()

        if not os.path.exists(fullfile):
            fp = open(fullfile, 'w')
            fp.write("[SFLvault]\n")
            fp.close()
            os.chmod(fullfile, 0o600)
        
        if not os.stat(fullfile)[0] & 0o600:
            # TODO: raise exception instead.
            print("Modes for %s must be 0600 (-rw-------)" % fullfile)
            sys.exit()

    def config_read(self):
        """Return the ConfigParser object, fully loaded"""    
        self.cfg = ConfigParser()
        fp = open(self.config_file, 'r')
        self.cfg.readfp(fp)
        fp.close()

        if not self.cfg.has_section('SFLvault'):
            self.cfg.add_section('SFLvault')

        if not self.cfg.has_section('Aliases'):
            self.cfg.add_section('Aliases')

        if not self.cfg.has_option('SFLvault', 'username'):
            self.cfg.set('SFLvault', 'username', '')
    
        if not self.cfg.has_option('SFLvault', 'url'):
            self.cfg.set('SFLvault', 'url', '')

    def config_write(self):
        """Write the ConfigParser element to disk."""
        fp = open(self.config_file, 'w')
        self.cfg.write(fp)
        fp.close()

    # Fake being the actual ConfigParser object
    def get(self, *args, **kwargs):
        return self.cfg.get(*args, **kwargs)
    def set(self, *args, **kwargs):
        return self.cfg.set(*args, **kwargs)
    def has_option(self, *args, **kwargs):
        return self.cfg.has_option(*args, **kwargs)

    def alias_add(self, alias, ptr):
        """Add an alias and save config."""
        tid = re.match(r'(.)#(\d+)', ptr)
        if not tid:
            raise ValueError("VaultID must be in the format: (.)#(\d+)")

        # Set the alias value
        self.cfg.set('Aliases', alias, ptr)
        # Save config.
        self.config_write()
        return True

    def alias_del(self, alias):
        """Remove an alias from the config.

        :rtype: True if removed, False otherwise.
        """
        if self.cfg.has_option('Aliases', alias):
            self.cfg.remove_option('Aliases', alias)
            self.config_write()
            return True
        else:
            return False

    def alias_list(self):
        """Return a list of aliases

        :rtype: list of (alias, value) pairs.
        """
        return self.cfg.items('Aliases')

    def alias_get(self, alias):
        """Return the pointer for a given alias"""
        if not self.cfg.has_option('Aliases', alias):
            return None
        else:
            return self.cfg.get('Aliases', alias)


    def _check_keyring(self):
        try:
            import keyring
        except ImportError as e:
            raise KeyringError("No keyring support, please install python-keyring")
        
    def wallet_list(self):
        """Return the list of available wallets, from Keyring

        [('0', 'Manual', None, 'Disabled', True),
         ('1', 'UncryptedKeyring', <UncryptedKeyring object>, 'Recommended', False),
         ...]
        """
        self._check_keyring()
        import keyring

        current = self.wallet_get()
        out = [('0', 'Manual', None, 'Disabled', current == None)]
        ref = {1: "Recommended", 0: "Supported", -1: "Not installed"} 
        for i, backend in enumerate(keyring.backend.get_all_keyring()):
            out.append((str(i + 1),
                        backend.__class__.__name__,
                        backend,
                        ref[backend.supported()],
                        backend.__class__.__name__ == current,
                        ))
        return out

    @property
    def _wallet_key_legacy(self):
        """Legacy scheme deriving a key from the sflvault config file path

It results in something like _'2ehome_2e$USERNAME_2esflvault_2econfig'
and breaks portability across machines using different usernames.
        """
        return re.subn(r'\.+', '.', re.subn(r'[-_:\\ /]', '.',
                                            self.config_file)[0].lower())[0]

    @property
    def _wallet_key(self):
        """Return a the key used to store the password in the keystore."""
        return 'password'

    def wallet_set(self, id, password):
        if id is None or id == '0':
            self.cfg.remove_option('SFLvault', 'wallet')
        else:
            lst = self.wallet_list()
            ids = [x[0] for x in lst]
            if id not in ids:
                raise KeyringError("No such Wallet ID: %s" % id)
            backend = [x for x in lst if x[0] == id][0]
            ret = backend[2].set_password("sflvault", self._wallet_key,
                                          password)
            if ret:  # Error saving the password ?
                raise KeyringError("Unable to store password in keyring")
            # Set the wallet value
            self.cfg.set('SFLvault', "wallet", backend[1])
        # Save config.
        self.config_write()
        return True

    def wallet_get(self):
        """Return the currently configured wallet, otherwise None"""
        if self.cfg.has_option('SFLvault', 'wallet'):
            return self.cfg.get('SFLvault', 'wallet')
        else:
            return None

    def wallet_get_obj(self):
        """Return the currently configured wallet as a backend object,
        otherwise None
        """
        name = self.wallet_get()
        if not name:
            return None

        lst = self.wallet_list()
        names = [x[1] for x in lst]
        if name not in names:
            raise KeyringError("No such Wallet type: %s" % name)
        self.wallet_test(name)

        backend = [x[2] for x in lst if x[1] == name][0]
        return backend

    def wallet_test(self, name):
        """Test if the Wallet is supported, otherwise, suggest to install
        packages
        """
        lst = self.wallet_list()
        backend = [x[2] for x in lst if x[1] == name][0]
        assoc = {'CryptedFileKeyring': 'python-keyring',
                 'UncryptedFileKeyring': 'python-keyring',
                 'KDEKWallet': 'python-keyring-kwallet',
                 'GnomeKeyring': 'python-keyring-gnome',
                 }
        if backend.supported() == -1:
            if name in assoc:
                add = ' To add support, please install the `%s` package.' % \
                    assoc[name]
            raise KeyringError("Keyring (%s) is not supported.%s "
                               "Use the `wallet` command to reconfigure." %
                               (name, add))
        return True

###
### On définit les fonctions qui vont traiter chaque sorte de requête.
###
class SFLvaultClient(object):
    """This is the main SFLvault Client object.

    It is used to script some access to the vault, to retrieve data, to store
    data, or to create a GUI interface on the top of it.

    Whether you want to access a local or remote Vault server, this is the
    object you need.
    """
    def __init__(self, config, shell=False):
        """Set up initial configuration for function calls

        :param config: Configuration filename to use.
        :param shell: if True, the private key will be cached for a while,
            not asking your password for each query to the vault.
        """
        # Load configuration
        self.cfg = SFLvaultConfig(config)

        # The function to call upon @authenticate to get passphrase from user.
        self.set_getpassfunc(None)

        self.shell_mode = shell
        self.authtok = ''
        self.authret = None
        # Set the default route to the Vault
        url = self.cfg.get('SFLvault', 'url')
        if url:
            self.vault = xmlrpc.client.Server(url, allow_none=True).sflvault

    def set_getpassfunc(self, func=None):
        """Set the function to ask for passphrase.

        By default, it is set to _getpass, which asks for the passphrase on the
        command line, but you can create a new function, that would for example
        pop-up a window, or use another mechanism to ask for passphrase and
        continue authentication."""
        if not func:
            self.getpassfunc = AskPassMethods(self.cfg).getpass
        else:
            self.getpassfunc = func
        
    def _set_vault(self, url, save=False):
        """Set the vault's URL and optionally save it"""
        self.vault = xmlrpc.client.Server(url, allow_none=True).sflvault
        if save:
            self.cfg.set('SFLvault', 'url', url)

    def vaultId(self, vid, prefix, check_alias=True):
        """Return an integer value for a given VaultID.
        
        A VaultID can be one of the following:
        
        * ``123`` - treated as is, and assumed to be of type `prefix`.
        * ``m#123`` - checked against `prefix`, otherwise raises an exception.
        * ``alias`` - checked against `prefix` and the aliases that are in
          the configuration, returns an integer, or raises an exception.

        :param check_alias: check for matching aliases if True, otherwise only
          the two first cases are treated.
        """
        #prefixes = ['m', 'u', 's', 'c'] # Machine, User, Service, Customer
        #if prefix not in prefixes:
        #    raise ValueError("Bad prefix for id %s (prefix given: %s)" % (id, prefix))
        
        # If it's only a numeric, assume it is of type 'prefix'.
        try:
            tmp = int(vid)
            return tmp
        except:
            pass

        # Match the m#123 formats..
        tid = re.match(r'(.)#(\d+)', vid)
        if tid:
            if tid.group(1) != prefix:
                raise VaultIDSpecError("Bad prefix for VaultID, "\
                                         "context requires '%s': %s"\
                                         % (prefix, vid))
            return int(tid.group(2))

        if check_alias:
            nid = self.cfg.alias_get(vid)

            if not nid:
                raise VaultIDSpecError("No such alias '%s'. Use `alias %s %s#[ID]` to set." % (vid, vid, prefix))

            return self.vaultId(nid, prefix, False)

        raise VaultIDSpecError("Invalid VaultID format: %s" % vid)



    ### REMOTE ACCESS METHODS


    @authenticate()
    def user_add(self, username, admin=False):
        # TODO: add support for --admin, to give admin privileges

        retval = vaultReply(self.vault.user_add(self.authtok, username, admin),
                            "Error adding user")

        print("Success: %s" % retval['message'])
        prms = (username, self.cfg.get('SFLvault', 'url'))
        print("The new user should run: sflvault user-setup %s %s" % prms)

        return retval


    @authenticate()
    def user_del(self, username):
        retval = vaultReply(self.vault.user_del(self.authtok, username),
                            "Error removing user")

        print("Success: %s" % retval['message'])

        return retval


    def _services_returned(self, retval):
        """Helper function for customer_del, machine_del and service_del."""
        
        if retval['error']:
            print("Error: %s" % retval['message'])

            if 'childs' in retval:
                print("Those services rely on services you were going "\
                      "to delete:")
                for x in retval['childs']:
                    print("     s#%s%s%s" % (x['id'],
                                             ' ' * (6 - len(str(x['id']))),
                                             x['url']))
        else:
            print("Success: %s" % retval['message'])


    @authenticate()
    def customer_del(self, customer_id):
        retval = self.vault.customer_del(self.authtok, customer_id)

        self._services_returned(retval)

        return retval
        

    @authenticate()
    def machine_del(self, machine_id):
        retval = self.vault.machine_del(self.authtok, machine_id)

        self._services_returned(retval)
        
        return retval


    @authenticate(True)
    def customer_get(self, customer_id):
        """Get information to be edited"""
        retval = vaultReply(self.vault.customer_get(self.authtok, customer_id),
                            "Error fetching data for customer %s" % customer_id)

        return retval['customer']

    @authenticate(True)
    def customer_put(self, customer_id, data):
        """Save the (potentially modified) customer to the Vault"""
        retval = vaultReply(self.vault.customer_put(self.authtok, customer_id,
                                                   data),
                            "Error saving data to vault, aborting.")

        print("Success: %s " % retval['message'])

        return retval
        

    @authenticate()
    def service_del(self, service_id):
        retval = self.vault.service_del(self.authtok, service_id)

        self._services_returned(retval)

        return retval


    @authenticate()
    def customer_add(self, customer_name):
        retval = vaultReply(self.vault.customer_add(self.authtok,
                                                    customer_name),
                            "Error adding customer")

        print("Success: %s" % retval['message'])
        print("New customer ID: c#%d" % retval['customer_id'])

        return retval


    @authenticate()
    def machine_add(self, customer_id, name='', fqdn='', ip='', location='', notes=''):
        """Add a machine to the database."""
        # customer_id REQUIRED
        retval = vaultReply(self.vault.machine_add(self.authtok,
                                                   int(customer_id),
                                                   name, fqdn,
                                                   ip, location,
                                                   notes),
                            "Error adding machine")
        print("Success: %s" % retval['message'])
        print("New machine ID: m#%d" % int(retval['machine_id']))

        return retval

    @authenticate()
    def service_add(self, machine_id, parent_service_id, url, group_ids, secret,
                    notes='', metadata=None):
        """Add a service to the Vault's database.

        Arguments:
            machine_id: A m#id machine identifier.
        
            parent_service_id: A s#id, parent service ID, to which you should connect before
            connecting to the service you're adding. Specify 0 or None if no parent exist. If you
            set this, machine_id is disregarded.
        
            url: URL of the service, with username, port and path if required
        
            group_ids: Multiple group IDs the service is part of. See `list-groups`
        
            notes: Simple text field, with notes.
        
            secret: Password for the service. Plain-text.
        
            metadata: Dictionary with metadata for services (depends on service).
        """

        # TODO: accept group_id as group_ids, accept list and send list.
        psi = int(parent_service_id) if parent_service_id else None
        retval = vaultReply(self.vault.service_add(self.authtok,
                                                  int(machine_id),
                                                  psi, url,
                                                  group_ids, secret,
                                                  notes,
                                                  metadata),
                            "Error adding service")

        print("Success: %s" % retval['message'])
        print("New service ID: s#%d" % retval['service_id'])

        return retval

    @authenticate()
    def service_passwd(self, service_id, newsecret):
        """Updates the password on the Vault for a certain service"""
        retval = vaultReply(self.vault.service_passwd(self.authtok,
                                                        service_id,
                                                        newsecret),
                            "Error changing password for "\
                            "service %s" % service_id)

        print("Success: %s" % retval['message'])
        print("Password updated for service: s#%d" % int(retval['service_id']))

        return retval
                            

    def _new_passphrase(self):
        """Return a new passphrase after asking arrogantly"""
        while True:
            passphrase = getpass.getpass("Enter passphrase (to secure "
                                         "your private key): ")
            passph2 = getpass.getpass("Enter passphrase again: ")

            if passphrase != passph2:
                print("Passphrase mismatch, try again.")
            elif passphrase == '':
                print("Passphrase cannot be null.")
            else:
                return passphrase

    def user_passwd(self, new_passphrase=None):
        """Change the password protecting the local private key."""
        old_passphrase = getpass.getpass('Enter your old passphrase: ')
        # Re-read configuration in case it changed since the shell is open.
        self.cfg.config_read()
        privkey_enc = self.cfg.get('SFLvault', 'key')
        # Decrypt the privkey
        try:
            thething = decrypt_privkey(privkey_enc, old_passphrase)
        except DecryptError as e:
            print("[SFLvault] Invalid passphrase")
            return
        del(privkey_enc)
        # Ask the new passphrase, 
        if not new_passphrase:
            new_passphrase = self._new_passphrase()
        # Encrypt and set the new passphrase
        self.cfg.set('SFLvault', 'key', encrypt_privkey(thething, new_passphrase))
        self.cfg.config_write()

        print("user-passwd successful")
        
    def user_setup(self, username, vault_url, passphrase=None):
        """Sets up the local configuration to communicate with the Vault.

        username  - the name with which an admin prepared (with add-user)
                    your account.
        vault_url - the URL pointing to the XML-RPC interface of the vault
                    (typically host://domain.example.org:5000/vault/rpc
        passphrase - use the given passphrase instead of asking it on the
                     command line.
        """
        # possible-TODO: implement --force if user wants to override.
        if self.cfg.has_option('SFLvault', 'key'):
            raise VaultConfigurationError("WARNING: you already have a private key stored in %s.  Backup/rename this file before running this command again." % (self.cfg.config_file))
            
        self._set_vault(vault_url, False)
        
        # Generate a new key:
        print("Generating new ElGamal key-pair...")
        eg = generate_elgamal_keypair()

        # Marshal the ElGamal key
        pubkey = elgamal_pubkey(eg)

        print("You will need a passphrase to secure your private key. The")
        print("encrypted key will be stored on this machine in %s" % self.cfg.config_file)
        print('-' * 80)

        if not passphrase:
            passphrase = self._new_passphrase()
        
        print("Sending request to vault...")
        # Send it to the vault, with username
        retval = vaultReply(self.vault.user_setup(username,
                                                serial_elgamal_pubkey(pubkey)),
                            "Setup failed")

        # If Vault sends a SUCCESS, save all the stuff (username, vault_url)
        # and encrypt privkey locally (with Blowfish)
        print("Vault says: %s" % retval['message'])

        # Save all (username, vault_url)
        # Encrypt privkey locally (with Blowfish)
        self.cfg.set('SFLvault', 'username', username)
        self._set_vault(vault_url, True)
        # p and x form the private key, add the public key, add g and y.
        # if encryption is required at some point.
        self.cfg.set('SFLvault', 'key',
                   encrypt_privkey(serial_elgamal_privkey(elgamal_bothkeys(eg)),
                                   passphrase))
        del(passphrase)
        del(eg)

        print("Saving settings...")
        self.cfg.config_write()
        return True


    @authenticate()
    def search(self, query, filters=None, verbose=True):
        """Search the database for query terms.

        Arguments:
            query: list of REGEXPs to be matched.
            
            filters: dict with keys in ['groups', 'machines', 'customers'] that limits the records
            returned to those matching those constraints. The values can be either int or str
            (representing an int).
            
            verbose (bool): shows the notes and location attributes for services and machines.

        Returns:
            Hierarchical view of the results.
        """

        # Remove empty filters:
        if filters:
            filters = dict([(x, filters[x]) for x in filters if filters[x]])

        retval = vaultReply(self.vault.search(self.authtok, query,
                 filters.get('groups') if filters else None, verbose, filters),
                            "Error searching database")
        print("Results:")

        # TODO: call the pager `less` when too long.
        level = 0
        for c_id, c in list(retval['results'].items()):
            level = 0
            # Display customer info
            print("c#%s  %s" % (c_id, c['name']))

            spc1 = ' ' * (4 + len(c_id))
            for m_id, m in list(c['machines'].items()):
                level = 1
                # Display machine infos: 
                add = ' ' * (4 + len(m_id))
                print("%sm#%s  %s (%s - %s)" % (spc1, m_id,
                                                m['name'],
                                                m['fqdn'], m['ip']))
                if verbose:
                    if m['location']:
                        print("%s%slocation: %s" % (spc1, add,
                                                 m['location']))
                    if m['notes']:
                        print("%s%snotes: %s" % (spc1, add,
                                                 m['notes']))

                spc2 = spc1 + add
                print("")
                for s_id, s in list(m['services'].items()):
                    level = 2
                    # Display service infos
                    add = ' ' * (4 + len(s_id))
                    p_id = s.get('parent_service_id')
                    print("%ss#%s  %s%s" % (spc2, s_id,
                                            s['url'],
                                            ("   (depends: s#%s)" % \
                                             p_id if p_id else '')))
                    if verbose:
                        if s['notes']:
                            print("%s%snotes: %s" % (spc2, add,
                                                     s['notes']))

                if level == 2:
                    print("%s" % (spc2) + '-' * (80 - len(spc2)))
                
            if level in [0,1]:
                print("%s" % (spc1) + '-' * (80 - len(spc1)))

        return retval
            
    def _decrypt_service(self, serv, onlysymkey=False, onlygroupkey=False):
        """Decrypt the service object returned from the vault.

        onlysymkey - return the plain symkey in the result
        onlygroupkey - return the plain groupkey ElGamal obj in result
        """
        # First decrypt groupkey
        try:
            # TODO: implement a groupkey cache system, since it's the longest
            #       thing to decrypt (over a second on a 3GHz machine)
            grouppacked = decrypt_longmsg(self.privkey, serv['cryptgroupkey'])
        except Exception as e:
            raise DecryptError("Unable to decrypt groupkey (%s)" % e)

        eg = ElGamal.ElGamalobj()
        (eg.p, eg.x, eg.g, eg.y) = unserial_elgamal_privkey(grouppacked)
        groupkey = eg
        
        if onlygroupkey:
            serv['groupkey'] = eg
            
        # Then decrypt symkey
        try:
            aeskey = decrypt_longmsg(groupkey, serv['cryptsymkey'])
        except Exception as e:
            raise DecryptError("Unable to decrypt symkey (%s)" % e)

        if onlysymkey:
            serv['symkey'] = aeskey

        if not onlygroupkey and not onlysymkey:
            serv['plaintext'] = decrypt_secret(aeskey, serv['secret'])


    @authenticate(True)
    def service_get(self, service_id, decrypt=True, group_id=None):
        """Get information to be edited"""
        if group_id:
            retval = vaultReply(self.vault.service_get(self.authtok,
                                                       service_id,
                                                       group_id),
                                "Error fetching data for service %s" % service_id)
        else:
            retval = vaultReply(self.vault.service_get(self.authtok, service_id),
                                "Error fetching data for service %s" % service_id)

        serv = retval['service']
        # Decrypt secret
        aeskey = ''
        secret = ''

        if decrypt:
            # Add it only if we can! (or if we want to)
            if serv.get('cryptgroupkey'):
                self._decrypt_service(serv)

        return serv


    @authenticate(True)
    def service_get_tree(self, service_id, with_groups=False):
        """Get information to be edited"""
        return self._service_get_tree(service_id, with_groups)

    def _service_get_tree(self, service_id, with_groups=False):
        """Same as service_get_tree, but without authentication, so
        that it can be called by ``show`` and ``connect``.
        """
        retval = vaultReply(self.vault.service_get_tree(self.authtok,
                                                        service_id,
                                                        with_groups),
                "Error fetching data-tree for service %s" % service_id)

        for x in retval['services']:
            # Decrypt secret
            aeskey = ''
            secret = ''

            if not x['cryptsymkey']:
                # Don't add a plaintext if we can't.
                continue

            self._decrypt_service(x)

        return retval['services']


    @authenticate(True)
    def service_put(self, service_id, data):
        """Save the (potentially modified) service to the Vault"""
        retval = vaultReply(self.vault.service_put(self.authtok, service_id,
                                                   data),
                            "Error saving data to vault, aborting.")

        print("Success: %s " % retval['message'])

        return retval
        

    @authenticate(True)
    def machine_get(self, machine_id):
        """Get information to be edited"""
        retval = vaultReply(self.vault.machine_get(self.authtok, machine_id),
                            "Error fetching data for machine %s" % machine_id)

        return retval['machine']

    @authenticate(True)
    def machine_put(self, machine_id, data):
        """Save the (potentially modified) machine to the Vault"""
        retval = vaultReply(self.vault.machine_put(self.authtok, machine_id,
                                                   data),
                            "Error saving data to vault, aborting.")

        print("Success: %s " % retval['message'])
        
        return retval

    @authenticate(True)
    def show(self, service_id, verbose=False, with_groups=False):
        """Show informations to connect to a particular service"""
        servs = self._service_get_tree(service_id, with_groups)
        self._show(servs, verbose)

    def _show(self, services, verbose=False):
        """Show results fetched by _service_get_tree, to be called by both
        `show` and `connect`."""
        print("Results:")
        pre = ''
        for x in services:
            # Show separator
            if pre:
                pass
                #print "%s%s" % (pre, '-' * (80-len(pre)))
                
            spc = len(str(x['id'])) * ' '

            secret = x['plaintext'] if 'plaintext' in x else '[access denied]'
            print("%ss#%d %s" % (pre, x['id'], x['url']))
            if x['groups_list']:
                groups = ', '.join(["g#%s %s" % (g[0], g[1])
                                    for g in x['groups_list']])
                print("%s%s   groups: %s" % (pre, spc, groups))
            print("%s%s   secret: %s" % (pre, spc, secret))
            
            if verbose:
                print("%s%s   notes: %s" % (pre,spc, x['notes']))
                if x['metadata']:
                    print("%s%s   metadata:" % (pre, spc))
                    for key, val in list(x['metadata'].items()):
                        print("%s%s     %s: %s" % (pre, spc, key, val))

            del(secret)

            pre = pre + '   ' + spc


    @authenticate(True)
    def connect(self, vid, with_show=False, command_line=''):
        """Connect to a distant machine (using SSH for now)"""
        servs = self._service_get_tree(vid)

        if with_show:
            self._show(servs)

        # Check and decrypt all ciphers prior to start connection,
        # if there are some missing, it's not useful to start.
        for x in servs:
            if not x['cryptsymkey']:
                raise RemotingError("We don't have access to password for service %s" % x['url'])

        connection = remoting.Chain(servs, command_line=command_line)
        connection.setup()
        connection.connect()
        #connection.debug_chain()

    @authenticate()
    def user_list(self, groups=False):
        """List users

        ``groups`` - if True, list groups for each user also
        """
        # Receive: [{'id': x.id, 'username': x.username,
        #            'created_time': x.created_time,
        #            'is_admin': x.is_admin,
        #            'setup_expired': x.setup_expired()}
        #            {}, {}, ...]
        #    
        retval = vaultReply(self.vault.user_list(self.authtok, groups),
                            "Error listing users")

        print("User list (with creation date):")
        
        to_clean = []  # Expired users to be removed
        for x in retval['list']:
            add = ''
            if x['is_admin']:
                add += ' [global admin]'
            if x['setup_expired']:
                add += ' [setup expired]'
                to_clean.append(x['username'])
            if x['waiting_setup'] and not x['setup_expired']:
                add += ' [in setup process]'

            # TODO: load the xmlrpclib.DateTime object into something more fun
            #       to deal with! Some day..
            print("u#%d\t%s\t%s %s" % (x['id'], x['username'],
                                       x['created_stamp'], add))

            if 'groups' in x:
                for grp in x['groups']:
                    add = ' [admin]' if grp['is_admin'] else ''
                    print("\t\tg#%s\t%s %s" % (grp['id'], grp['name'], add))

        print('-' * 80)

        if len(to_clean):
            print("There are expired users. To remove them, run:")
            for usr in to_clean:
                print("   sflvault user-del -u %s" % usr)

        return retval
        

    @authenticate(True)
    def group_get(self, group_id):
        """Get information to be edited"""
        retval = vaultReply(self.vault.group_get(self.authtok, group_id),
                            "Error fetching data for group %s" % group_id)

        return retval['group']

    @authenticate(True)
    def group_put(self, group_id, data):
        """Save the (potentially modified) Group to the Vault"""
        retval = vaultReply(self.vault.group_put(self.authtok, group_id,
                                                   data),
                            "Error saving data to vault, aborting.")

        print("Success: %s " % retval['message'])
        
        return retval


    @authenticate(True)
    def group_add_service(self, group_id, service_id):
        print("Fetching service info...")
        retval = vaultReply(self.vault.service_get(self.authtok, service_id),
                            "Error loading service infos")

        # TODO: decrypt the symkey with the group's decrypted privkey.
        serv = retval['service']

        print("Decrypting symkey...")
        self._decrypt_service(serv, onlysymkey=True)


        print("Sending data back to vault")
        retval = vaultReply(self.vault.group_add_service(self.authtok,
                                                         group_id,
                                                         service_id,
                                                         serv['symkey']),
                            "Error adding service to group")

        print("Success: %s" % retval['message'])

        return retval

    @authenticate()
    def group_del_service(self, group_id, service_id):
        retval = vaultReply(self.vault.group_del_service(self.authtok, group_id,
                                                 service_id),
                            "Error removing service from group")

        print("Success: %s" % retval['message'])

        return retval

    @authenticate(True)
    def group_add_user(self, group_id, user, is_admin=False):
        retval = vaultReply(self.vault.group_add_user(self.authtok, group_id,
                                                      user),
                            "Error adding user to group")

        # Decrypt cryptgroupkey
        # TODO: make use of a cache
        grouppacked = decrypt_longmsg(self.privkey, retval['cryptgroupkey'])
        
        # Get userpubkey and unpack
        eg = ElGamal.ElGamalobj()
        (eg.p, eg.g, eg.y) = unserial_elgamal_pubkey(retval['userpubkey'])
        
        # Re-encrypt for user
        newcryptgroupkey = encrypt_longmsg(eg, grouppacked)
        
        # Return a well-formed database-ready cryptgroupkey for user,
        # also, give the param is_admin.. as desired.
        retval = vaultReply(self.vault.group_add_user(self.authtok, group_id,
                                                      user, is_admin,
                                                      newcryptgroupkey),
                            "Error adding user to group")

        print("Success: %s" % retval['message'])

        return retval

    @authenticate()
    def group_del_user(self, group_id, user):
        retval = vaultReply(self.vault.group_del_user(self.authtok, group_id,
                                                      user),
                            "Error removing user from group")

        print("Success: %s" % retval['message'])
        return retval
    
    @authenticate()
    def group_add(self, group_name):
        """Add a named group to the Vault. Return the group id."""
        
        print("Please wait, Vault generating keypair...")
        
        retval = vaultReply(self.vault.group_add(self.authtok, group_name),
                            "Error adding group")

        print("Success: %s " % retval['message'])
        print("New group id: g#%d" % retval['group_id'])

        return retval


    @authenticate()
    def group_del(self, group_id, delete_cascade=False):
        """Remove a group from the Vault, making sure no services are left
        behind."""
        retval = vaultReply(self.vault.group_del(self.authtok, group_id, delete_cascade),
                            "Error removing group")

        print("Success: %s" % retval['message'])
        return retval

    @authenticate()
    def group_list(self, quiet=False):
        """Simply list the available groups"""
        retval = vaultReply(self.vault.group_list(self.authtok),
                            "Error listing groups")

        print("Groups:")

        lst = retval['list']
        lst.sort(key=lambda x: x['name'])
        
        show_legend = False
        for grp in lst:
            add = []
            if grp.get('hidden', False):
                add.append('[hidden]')
            if grp.get('member', False):
                add.append('[member]')
            if grp.get('admin', False):
                add.append('[admin]')
            print("\tg#%d\t%s %s" % (grp['id'], grp['name'], ' '.join(add)))

            if not quiet and 'members' in grp:
                show_legend = any(bool(x[2]) for x in grp['members']) or \
                    show_legend
                l = ["%s%s" % ('*' if x[2] else '', x[1])
                     for x in grp['members']]
                print("\t\tMembers: %s" % ', '.join(l))
        if not quiet and show_legend:
            print("* = group admin")
        return retval


    @authenticate()
    def machine_list(self, verbose=False, customer_id=None):
        retval = vaultReply(self.vault.machine_list(self.authtok, customer_id),
                            "Error listing machines")

        print("Machines list:")

        oldcid = 0
        for x in retval['list']:
            if oldcid != x['customer_id']:
                print("%s (c#%d)" % (x['customer_name'], x['customer_id']))
                oldcid = x['customer_id']
            print("\tm#%d\t%s (%s)" % (x['id'], x['name'], x['fqdn'] or x['ip']))
            if verbose:
                print("\t\tLocation: %s" % x['location'].replace('\n', '\t\t\n'))
                print("\t\tNotes: %s" % x['notes'].replace('\n', '\t\t\n'))
                print('-' * 76)

        return retval


    @authenticate()
    def customer_list(self, customer_id=None):
        """List customers in the vault and possibly corresponding to the needed id
        
        Keywords arguments:
            customer_id -- Id of the needed customer to list

        Returns a list of dicts:
            [{'id': '%d','name': 'blah'}]

        """
        retval = vaultReply(self.vault.customer_list(self.authtok),
                            "Error listing customers")
        print("Customer list:")
        for x in retval['list']:
            print("c#%d\t%s" % (x['id'], x['name']))
        return retval





# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008-2021  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    from setuptools import setup, find_packages
except ImportError:
    from distribute_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

import platform


setup(
    name='SFLvault-client',
    version="0.9.1",
    description='Networked credentials store and authentication manager - Client',
    author='Savoir-faire Linux',
    author_email='contact@savoirfairelinux.com',
    url='https://www.sflvault.org',
    license='GPLv3',
    install_requires=["SFLvault-common~=0.9",
                      # TODO: Port to a recent keyring.  This old 1.6
                      # version works until Python 3.8 (see:
                      # https://github.com/savoirfairelinux/sflvault/issues/51).
                      "keyring~=1.6",
                      "pycrypto",
                      "decorator",
                      ] + \
                     ([] if platform.system() == 'Windows'
                         else ["urwid>=0.9.8.1",
                               "pexpect>=2.5"]),
    packages=find_packages(),
    namespace_packages=['sflvault'],
    include_package_data=True,
    test_suite='nose.collector',
    package_data={'sflvault': ['i18n/*/LC_MESSAGES/*.mo']},
    #message_extractors = {'sflvault': [
    #        ('**.py', 'python', None),
    #        ('templates/**.mako', 'mako', None),
    #        ('public/**', 'ignore', None)]},
    entry_points="""
    [console_scripts]
    sflvault = sflvault.client.commands:main

    [sflvault.services]
    ssh = sflvault.client.services:ssh
    ssh+pki = sflvault.client.services:ssh_pki
    content = sflvault.client.services:content
    sflvault = sflvault.client.services:sflvault
    vnc = sflvault.client.services:vnc
    mysql = sflvault.client.services:mysql
    psql = sflvault.client.services:postgres
    postgres = sflvault.client.services:postgres
    postgresql = sflvault.client.services:postgres
    su = sflvault.client.services:su
    sudo = sflvault.client.services:sudo

    """,
)



# -=- encoding: utf-8 -=-
#
# SFLvault - Secure networked password store and credentials manager.
#
# Copyright (C) 2008  Savoir-faire Linux inc.
#
# Author: Alexandre Bourget <alexandre.bourget@savoirfairelinux.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Several utilities used in the SFLvault client"""

import urllib.parse
import re
from pkg_resources import iter_entry_points, DistributionNotFound
import platform
if platform.system() != 'Windows':
    import readline


__all__ = ['urllib', 'AuthenticationError', 'PermissionError',
           'VaultIDSpecError', 'VaultConfigurationError', 'RemotingError',
           'ServiceRequireError', 'ServiceExpectError', 'sflvault_escape_chr',
           'ask_for_service_password', 'services_entry_points',
           "ServiceSwitchException", "KeyringError"]


def services_entry_points():
    """Return the list of entry points for the different services."""
    return iter_entry_points('sflvault.services')

#
# Add protocols to urlparse, for correct parsing of ssh and others.
#
urllib.parse.uses_netloc.extend(['ssh', 'vlc', 'vpn', 'openvpn', 'git',
                             'bzr+ssh', 'vnc', 'mysql', 'sudo', 'su',
                             'psql'] +
                             [x.name for x in services_entry_points()])


# Issue: Ctrl+Alt+;
sflvault_escape_chr = chr(30)

def ask_for_service_password(edit=False, url=None):
    # Use the module's ask_password if it supports one...
    parsed_url = urllib.parse.urlparse(url)
    for ep in services_entry_points():
        if ep.name == parsed_url.scheme:
            try:
                srvobj = ep.load()
            except DistributionNotFound as e:
                break
            if hasattr(srvobj, 'ask_password'):
                return srvobj.ask_password(edit, parsed_url)

    # Use raw_input so that we see the password. To make sure we enter
    # a valid and the one we want (what if copy&paste didn't work, and
    # you didn't know ?)
    sec = input("Enter new service's password: " if edit
                    else "Enter service's password: ")
    if readline.get_current_history_length() > 0:
        readline.remove_history_item(readline.get_current_history_length() - 1)
    # CSI n F to go back one line. CSI n K to erase the line.
    # See http://en.wikipedia.org/wiki/ANSI_escape_code
    print("\033[1F\033[2K ... password taken ...")
    return sec


### Passwords management exceptions

class KeyringError(Exception):
    pass



### Authentication Exceptions

class AuthenticationError(Exception):
    pass


### Server restrictions errors

class PermissionError(Exception):
    """Raised when server enforces permissions"""
    pass


### VaultError is in SFLvault-common


### Vault-communication Exceptions
    
class VaultIDSpecError(Exception):
    """When bad parameters are passed to vaultId"""
    pass

class VaultConfigurationError(Exception):
    """Except when we're missing some config info."""
    pass

### Remoting Exceptions

class RemotingError(Exception):
    """When something happens in the Remoting mechanisms"""
    pass

class ServiceRequireError(RemotingError):
    """When the required() elements can't do what we want."""
    pass

class ServiceExpectError(RemotingError):
    """When an error occurs in the expected values in prework to set up a
    service.

    We assume the interact() will be called for the parent at that point."""
    pass

class ServiceSwitchException(Exception):
    """Way to switch the active shell"""
    pass


# Flexible URL parser (not RFC compliant, but very flexible for our purpose)
class URLParserError(Exception):
    pass

class URLParser(object):
    """Parses URLs and splits `scheme`, `hostname`, `username`, `port`,
    `path`, `query`, `fragment`, etc.. apart in an string representing an URL

    >>> u = URLParser('http://www.example.com/path')
    >>> u
    ...
    >>> u.hostname
    'www.example.com'
    >>> u.path
    '/path'

    >>> u2 = URLParser('git+ssh://[user@host]@git.example.com/var/www/repos')
    >>> u2.username
    'user@host'
    >>> u2.hostname
    'git.example.com'
    >>> u2.path
    '/var/www/repos'

    >>> u3 = URLParser('https://[user@host]:passwd123@[2009::10:ab]:123/var/my/path?q=hello#frag123')
    >>> u3.scheme
    'https'
    >>> u3.username
    'user@host'
    >>> u3.password
    'passwd123'
    >>> u3.hostname
    '2009::10:ab'
    >>> u3.port
    '123'
    >>> u3.path
    '/var/my/path'
    >>> u3.query
    'q=hello'
    >>> u3.fragment
    'frag123'

    >>> u3.gen_url(with_password=False)  # Default behavior
    'https://[user@host]@[2009::10:ab]:123/var/my/path?q=hello#frag123'
    """
    _regex = re.compile(r"([a-zA-Z0-1+-]+)(://)(((\[([^\]]+)\])|([^@]+))(:([^@]*))?@)?(([^\[/][^:/\?#]+)|(\[([^\]]+)\]))?(:(\d+))?(/([^\?]*))?(\?([^#]*))?(#(.*))?")
    
    def __init__(self, url=None):
        """Parse an URL or create a new empty object"""
        self.scheme = None
        self.username = None
        self.password = None
        self.hostname = None
        self.port = None
        self.path = None
        self.query = None
        self.fragment = None
        if url:
            self._parse(url)

    def _parse(self, url):
        """Extract infos from URLs, and fill in this object"""
        res = self._regex.match(url)
        if not res:
            raise URLParserError('Invalid or malformed URL')
        self.scheme = res.group(1)
        self.username = res.group(6) or res.group(7) or ''
        self.password = res.group(9)
        self.hostname = res.group(11)or res.group(13) or ''
        self.port = res.group(15)
        self.path = '/' + (res.group(17) or '')
        self.query = res.group(19) or ''
        self.fragment = res.group(21) or ''
        self.res = res

    def _show(self):
        for i in range(len(self.res.groups())):
            print(i+1, self.res.group(i+1))

    def gen_url(self, with_password=False):
        """Renders the URL, optionally without the password, based on
        the internal state of it's attributes (hostname, username, etc..)

        >>> u = URLParser('http://www.example.com/path')
        >>> u.
        """
        if not self.scheme:
            raise URLParserError('No scheme specified, please set the `scheme`'
                                 ' attribute')
        s = ['%s://' % self.scheme]
        if self.username:
            u = '[%s]' if set('@:?/#') & set(self.username) else '%s'
            if with_password and self.password:
                u += ':%s' % self.password
            s += ["%s@" % (u % self.username)]
        if self.hostname:
            h = '[%s]' if set(r':?/#') & set(self.hostname) else '%s'
            s += [h % self.hostname]
        if self.port:
            s += [':%s' % self.port]
        if self.path:
            s += [self.path]
        if self.query:
            s += ['?%s' % self.query]
        if self.fragment:
            s += ['#%s' % self.fragment]
        return ''.join(s)

    def __repr__(self):
        return '<URLParser: %s (passwd: %s)>' % (self.gen_url(), self.password)
